import os

from twisted.internet.defer import Deferred

from landscape.package.reporter import find_reporter_command
from landscape.package.store import PackageStore

from landscape.monitor.packagemonitor import PackageMonitor
from landscape.tests.helpers import (
    LandscapeTest, EnvironSaverHelper, MonitorHelper)


class PackageMonitorTest(LandscapeTest):
    """Tests for the temperature plugin."""

    helpers = [EnvironSaverHelper, MonitorHelper]

    def setUp(self):
        """Initialize test helpers and create a sample thermal zone."""
        super(PackageMonitorTest, self).setUp()
        self.package_store_filename = self.makeFile()
        self.package_store = PackageStore(self.package_store_filename)

        self.package_monitor = PackageMonitor(self.package_store_filename)

    def createReporterTask(self):
        """
        Put a task for the package reported into the package store.
        """
        message = {"type": "package-ids", "ids": [None], "request-id": 1}
        return self.package_store.add_task("reporter", message)

    def assertSingleReporterTask(self, data, task_id):
        """
        Check that we have exactly one task, that it contains the right data
        and that it's ID matches our expectation.
        """
        # The next task should contain the passed data.
        task = self.package_store.get_next_task("reporter")
        self.assertEqual(task.data, data)

        # We want to make sure it has the correct id of 2 so that we
        # know it's not a new task that the reporter could possibly
        # remove by accident.
        self.assertEqual(task.id, task_id)

        # Let's remove that task and make sure there are no more tasks
        # in the queue.
        task.remove()
        task = self.package_store.get_next_task("reporter")
        self.assertEqual(task, None)

    def test_create_default_store_upon_message_handling(self):
        """
        If the package sqlite database file doesn't exist yet, it is created
        upon message handling.
        """
        filename = os.path.join(self.broker_service.config.data_path,
                                "package/database")
        package_monitor = PackageMonitor()
        os.unlink(filename)
        self.assertFalse(os.path.isfile(filename))

        self.monitor.add(package_monitor)
        package_monitor_mock = self.mocker.patch(package_monitor)
        package_monitor_mock.spawn_reporter()
        self.mocker.replay()

        message = {"type": "package-ids"}
        self.monitor.dispatch_message(message)
        self.assertTrue(os.path.isfile(filename))

    def test_run_interval(self):
        """
        The C{run_interval} of L{PackageMonitor} can be customized via the
        C{package_monitor_interval} configuration parameter.
        """
        self.monitor.config.package_monitor_interval = 1234
        self.package_monitor.register(self.monitor)
        self.assertEqual(1234, self.package_monitor.run_interval)

    def test_dont_spawn_reporter_if_message_not_accepted(self):
        self.monitor.add(self.package_monitor)

        package_monitor_mock = self.mocker.patch(self.package_monitor)
        package_monitor_mock.spawn_reporter()
        self.mocker.count(0)

        self.mocker.replay()

        return self.package_monitor.run()

    def test_spawn_reporter_on_registration_when_already_accepted(self):
        package_monitor_mock = self.mocker.patch(self.package_monitor)
        package_monitor_mock.spawn_reporter()

        # Slightly tricky as we have to wait for the result of run(),
        # but we don't have its deferred yet.  To handle it, we create
        # our own deferred, and register a callback for when run()
        # returns, chaining both deferreds at that point.
        deferred = Deferred()

        def run_has_run(run_result_deferred):
            return run_result_deferred.chainDeferred(deferred)

        package_monitor_mock.run()
        self.mocker.passthrough(run_has_run)

        self.mocker.replay()

        self.broker_service.message_store.set_accepted_types(["packages"])
        self.monitor.add(self.package_monitor)

        return deferred

    def test_spawn_reporter_on_run_if_message_accepted(self):

        self.broker_service.message_store.set_accepted_types(["packages"])

        package_monitor_mock = self.mocker.patch(self.package_monitor)
        package_monitor_mock.spawn_reporter()
        self.mocker.count(2)  # Once for registration, then again explicitly.
        self.mocker.replay()

        self.monitor.add(self.package_monitor)
        return self.package_monitor.run()

    def test_package_ids_handling(self):
        self.monitor.add(self.package_monitor)

        package_monitor_mock = self.mocker.patch(self.package_monitor)
        package_monitor_mock.spawn_reporter()
        self.mocker.replay()

        message = {"type": "package-ids", "ids": [None], "request-id": 1}
        self.monitor.dispatch_message(message)
        task = self.package_store.get_next_task("reporter")
        self.assertTrue(task)
        self.assertEqual(task.data, message)

    def test_spawn_reporter(self):
        command = self.makeFile("#!/bin/sh\necho 'I am the reporter!' >&2\n")
        os.chmod(command, 0755)
        find_command_mock = self.mocker.replace(find_reporter_command)
        find_command_mock()
        self.mocker.result(command)
        self.mocker.replay()

        package_monitor = PackageMonitor(self.package_store_filename)
        self.monitor.add(package_monitor)
        result = package_monitor.spawn_reporter()

        def got_result(result):
            log = self.logfile.getvalue()
            self.assertIn("I am the reporter!", log)
            self.assertNotIn(command, log)

        return result.addCallback(got_result)

    def test_spawn_reporter_without_output(self):
        find_command_mock = self.mocker.replace(find_reporter_command)
        find_command_mock()
        self.mocker.result("/bin/true")
        self.mocker.replay()

        package_monitor = PackageMonitor(self.package_store_filename)
        self.monitor.add(package_monitor)
        result = package_monitor.spawn_reporter()

        def got_result(result):
            log = self.logfile.getvalue()
            self.assertNotIn("reporter output", log)

        return result.addCallback(got_result)

    def test_spawn_reporter_copies_environment(self):
        command = self.makeFile("#!/bin/sh\necho VAR: $VAR\n")
        os.chmod(command, 0755)
        find_command_mock = self.mocker.replace(find_reporter_command)
        find_command_mock()
        self.mocker.result(command)
        self.mocker.replay()

        package_monitor = PackageMonitor(self.package_store_filename)
        self.monitor.add(package_monitor)

        os.environ["VAR"] = "HI!"

        result = package_monitor.spawn_reporter()

        def got_result(result):
            log = self.logfile.getvalue()
            self.assertIn("VAR: HI!", log)
            self.assertNotIn(command, log)

        return result.addCallback(got_result)

    def test_spawn_reporter_passes_quiet_option(self):
        command = self.makeFile("#!/bin/sh\necho OPTIONS: $@\n")
        os.chmod(command, 0755)
        find_command_mock = self.mocker.replace(find_reporter_command)
        find_command_mock()
        self.mocker.result(command)
        self.mocker.replay()

        package_monitor = PackageMonitor(self.package_store_filename)
        self.monitor.add(package_monitor)

        result = package_monitor.spawn_reporter()

        def got_result(result):
            log = self.logfile.getvalue()
            self.assertIn("OPTIONS: --quiet", log)
            self.assertNotIn(command, log)

        return result.addCallback(got_result)

    def test_call_on_accepted(self):
        package_monitor_mock = self.mocker.patch(self.package_monitor)
        package_monitor_mock.spawn_reporter()

        self.mocker.replay()

        self.monitor.add(self.package_monitor)
        self.monitor.reactor.fire(
            ("message-type-acceptance-changed", "packages"), True)

    def test_resynchronize(self):
        """
        If a 'resynchronize' reactor event is fired with 'package' scope, the
        package monitor should clear all queued tasks and queue a task that
        tells the report to clear out the rest of the package data.
        """
        self.monitor.add(self.package_monitor)
        self.createReporterTask()

        # The server doesn't currently send 'package' scope, but we should
        # support it in case we change that in the future.
        package_scope = ["package"]
        self.monitor.reactor.fire("resynchronize", package_scope)

        self.assertSingleReporterTask({"type": "resynchronize"}, 2)

    def test_resynchronize_gets_new_session_id(self):
        """
        When a 'resynchronize' reactor event is fired, the C{PackageMonitor}
        acquires a new session ID (as the old one will be blocked).
        """
        self.monitor.add(self.package_monitor)
        session_id = self.package_monitor._session_id
        self.createReporterTask()

        self.package_monitor.client.broker.message_store.drop_session_ids()
        self.monitor.reactor.fire("resynchronize")
        self.assertNotEqual(session_id, self.package_monitor._session_id)

    def test_resynchronize_on_global_scope(self):
        """
        If a 'resynchronize' reactor event is fired with global scope (the
        empty list) , the package monitor should act as if it were an event
        with 'package' scope.
        """
        self.monitor.add(self.package_monitor)
        self.createReporterTask()

        self.monitor.reactor.fire("resynchronize")

        # The next task should be the resynchronize message.
        self.assertSingleReporterTask({"type": "resynchronize"}, 2)

    def test_not_resynchronize_with_other_scope(self):
        """
        If a 'resynchronize' reactor event is fired with an irrelevant scope,
        the package monitor should not respond to this.
        """
        self.monitor.add(self.package_monitor)
        task = self.createReporterTask()

        disk_scope = ["disk"]
        self.monitor.reactor.fire("resynchronize", disk_scope)

        # The next task should *not* be the resynchronize message, but instead
        # the original task we created.
        self.assertSingleReporterTask(task.data, task.id)

    def test_spawn_reporter_doesnt_chdir(self):
        command = self.makeFile("#!/bin/sh\necho RUN\n")
        os.chmod(command, 0755)
        cwd = os.getcwd()
        self.addCleanup(os.chdir, cwd)
        dir = self.makeDir()
        os.chdir(dir)
        os.chmod(dir, 0)

        find_command_mock = self.mocker.replace(find_reporter_command)
        find_command_mock()
        self.mocker.result(command)
        self.mocker.replay()

        package_monitor = PackageMonitor(self.package_store_filename)
        self.monitor.add(package_monitor)

        result = package_monitor.spawn_reporter()

        def got_result(result):
            log = self.logfile.getvalue()
            self.assertIn("RUN", log)
            # restore permissions to the dir so tearDown can clean it up
            os.chmod(dir, 0766)

        return result.addCallback(got_result)

    def test_changing_server_uuid_clears_hash_ids(self):
        """
        The package hash=>id map is server-specific, so when we change
        servers, we should reset this map.
        """
        self.package_store.set_hash_ids({"hash1": 1, "hash2": 2})
        self.monitor.add(self.package_monitor)
        self.monitor.reactor.fire("server-uuid-changed", "old", "new")

        self.assertEqual(self.package_store.get_hash_id("hash1"), None)
        self.assertEqual(self.package_store.get_hash_id("hash2"), None)

    def test_changing_server_uuid_wont_clear_hash_ids_with_old_uuid_none(self):
        """
        If the old UUID is unknown, that means the client just started
        talking to a server that knows how to communicate its UUID, so we
        don't want to clear the old hashes in this case.
        """
        self.package_store.set_hash_ids({"hash1": 1, "hash2": 2})
        self.monitor.add(self.package_monitor)
        self.monitor.reactor.fire("server-uuid-changed", None, "new-uuid")
        self.assertEqual(self.package_store.get_hash_id("hash1"), 1)
        self.assertEqual(self.package_store.get_hash_id("hash2"), 2)



from twisted.internet import reactor
from twisted.internet.error import ConnectError, ConnectionDone
from twisted.internet.task import Clock
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.python.failure import Failure

from landscape.lib.amp import (
    MethodCallError, MethodCallServerProtocol, MethodCallClientProtocol,
    MethodCallServerFactory, MethodCallClientFactory, RemoteObject,
    MethodCallSender)
from landscape.tests.helpers import LandscapeTest


class FakeTransport(object):
    """Accumulate written data into a list."""

    def __init__(self, connection):
        self.stream = []
        self.connection = connection

    def write(self, data):
        self.stream.append(data)

    def loseConnection(self):
        raise NotImplemented()

    def getPeer(self):
        pass

    def getHost(self):
        pass


class FakeConnection(object):
    """Simulate a connection between a client and a server protocol."""

    def __init__(self, client, server):
        self.client = client
        self.server = server

    def make(self):
        self.server.makeConnection(FakeTransport(self))
        self.client.makeConnection(FakeTransport(self))

    def lose(self, connector, reason):
        self.server.connectionLost(reason)
        self.client.connectionLost(reason)
        self.client.factory.clientConnectionLost(connector, reason)

    def flush(self):
        """
        Notify the server of any data written by the client and viceversa.
        """
        while True:
            if self.client.transport and self.client.transport.stream:
                self.server.dataReceived(self.client.transport.stream.pop(0))
            elif self.server.transport and self.server.transport.stream:
                self.client.dataReceived(self.server.transport.stream.pop(0))
            else:
                break


class FakeConnector(object):
    """Make L{FakeConnection}s using the given server and client factories."""

    def __init__(self, client, server):
        self.client = client
        self.server = server
        self.connection = None

    @property
    def factory(self):
        return self.client

    def connect(self):
        self.connection = FakeConnection(self.client.buildProtocol(None),
                                         self.server.buildProtocol(None))

        # XXX Let the client factory be aware of this fake connection, so
        # it can flush it when needed. This is to workaround AMP not
        # supporting synchronous transports
        self.client.fake_connection = self.connection

        self.connection.make()

    def disconnect(self):
        self.connection.lose(self, Failure(ConnectionDone()))


class DummyObject(object):

    method = None


class MethodCallTest(LandscapeTest):

    def setUp(self):
        super(MethodCallTest, self).setUp()
        self.methods = ["method"]
        self.object = DummyObject()
        server = MethodCallServerProtocol(self.object, self.methods)
        client = MethodCallClientProtocol()
        self.connection = FakeConnection(client, server)
        self.connection.make()
        self.clock = Clock()
        self.sender = MethodCallSender(client, self.clock)

    def test_with_forbidden_method(self):
        """
        If a method is not included in L{MethodCallServerFactory.methods} it
        can't be called.
        """
        self.methods.remove("method")
        deferred = self.sender.send_method_call(method="method",
                                                args=[],
                                                kwargs={})
        self.connection.flush()
        self.failureResultOf(deferred).trap(MethodCallError)

    def test_with_no_arguments(self):
        """
        A connected client can issue a L{MethodCall} without arguments and
        with an empty response.
        """
        self.object.method = lambda: None
        deferred = self.sender.send_method_call(method="method",
                                                args=[],
                                                kwargs={})
        self.connection.flush()
        self.assertIs(None, self.successResultOf(deferred))

    def test_with_return_value(self):
        """
        A connected client can issue a L{MethodCall} targeted to an
        object method with a return value.
        """
        self.object.method = lambda: "Cool result"
        deferred = self.sender.send_method_call(method="method",
                                                args=[],
                                                kwargs={})
        self.connection.flush()
        self.assertEqual("Cool result", self.successResultOf(deferred))

    def test_with_one_argument(self):
        """
        A connected AMP client can issue a L{MethodCall} with one argument and
        a response value.
        """
        self.object.method = lambda word: word.capitalize()
        deferred = self.sender.send_method_call(method="method",
                                                args=["john"],
                                                kwargs={})
        self.connection.flush()
        self.assertEqual("John", self.successResultOf(deferred))

    def test_with_boolean_return_value(self):
        """
        The return value of a L{MethodCall} argument can be a boolean.
        """
        self.object.method = lambda word: len(word) < 3
        deferred = self.sender.send_method_call(method="method",
                                                args=["hi"],
                                                kwargs={})
        self.connection.flush()
        self.assertTrue(self.successResultOf(deferred))

    def test_with_many_arguments(self):
        """
        A connected client can issue a L{MethodCall} with many arguments.
        """
        self.object.method = lambda word1, word2: word1 + word2
        deferred = self.sender.send_method_call(method="method",
                                                args=["We ", "rock"],
                                                kwargs={})
        self.connection.flush()
        self.assertEqual("We rock", self.successResultOf(deferred))

    def test_with_default_arguments(self):
        """
        A connected client can issue a L{MethodCall} for methods having
        default arguments.
        """
        self.object.method = lambda word, index=0: word[index:].lower()
        deferred = self.sender.send_method_call(method="method",
                                                args=["OHH"],
                                                kwargs={})
        self.connection.flush()
        self.assertEqual("ohh", self.successResultOf(deferred))

    def test_with_overriden_default_arguments(self):
        """
        A connected client can issue a L{MethodCall} with keyword arguments
        having default values in the target object.  If a value is specified by
        the caller it will be used in place of the default value
        """
        self.object.method = lambda word, index=0: word[index:].lower()
        deferred = self.sender.send_method_call(method="method",
                                                args=["ABC"],
                                                kwargs={"index": 2})
        self.connection.flush()
        self.assertEqual("c", self.successResultOf(deferred))

    def test_with_dictionary_arguments(self):
        """
        Method arguments passed to a L{MethodCall} can be dictionaries.
        """
        self.object.method = lambda d: "".join(d.keys()) * sum(d.values())
        deferred = self.sender.send_method_call(method="method",
                                                args=[{"foo": 1, "bar": 2}],
                                                kwargs={})
        self.connection.flush()
        self.assertEqual("foobarfoobarfoobar", self.successResultOf(deferred))

    def test_with_non_serializable_return_value(self):
        """
        If the target object method returns an object that can't be serialized,
        the L{MethodCall} raises an error.
        """
        class Complex(object):
            pass

        self.object.method = lambda: Complex()
        deferred = self.sender.send_method_call(method="method",
                                                args=[],
                                                kwargs={})
        self.connection.flush()
        self.failureResultOf(deferred).trap(MethodCallError)

    def test_with_long_argument(self):
        """
        The L{MethodCall} protocol supports sending method calls with arguments
        bigger than the maximum AMP parameter value size.
        """
        self.object.method = lambda word: len(word) == 65535
        deferred = self.sender.send_method_call(method="method",
                                                args=["!" * 65535],
                                                kwargs={})
        self.connection.flush()
        self.assertTrue(self.successResultOf(deferred))

    def test_with_long_argument_multiple_calls(self):
        """
        The L{MethodCall} protocol supports concurrently sending multiple
        method calls with arguments bigger than the maximum AMP value size.
        """
        self.object.method = lambda word: len(word)
        deferred1 = self.sender.send_method_call(method="method",
                                                 args=["!" * 80000],
                                                 kwargs={})
        deferred2 = self.sender.send_method_call(method="method",
                                                 args=["*" * 90000],
                                                 kwargs={})

        self.connection.flush()
        self.assertEqual(80000, self.successResultOf(deferred1))
        self.assertEqual(90000, self.successResultOf(deferred2))

    def test_with_exception(self):
        """
        If the target object method raises an exception, the remote call fails
        with a L{MethodCallError}.
        """
        self.object.method = lambda a, b: a / b
        deferred = self.sender.send_method_call(method="method",
                                                args=[1, 0],
                                                kwargs={})
        self.connection.flush()
        self.failureResultOf(deferred).trap(MethodCallError)

    def test_with_successful_deferred(self):
        """
        If the target object method returns a L{Deferred}, it is handled
        transparently.
        """
        self.object.deferred = Deferred()
        self.object.method = lambda: self.object.deferred
        result = []
        deferred = self.sender.send_method_call(method="method",
                                                args=[],
                                                kwargs={})
        deferred.addCallback(result.append)

        self.connection.flush()

        # At this point the receiver is waiting for method to complete, so
        # the deferred has not fired yet
        self.assertEqual([], result)

        # Fire the deferred and let the receiver respond
        self.object.deferred.callback("Hey!")
        self.connection.flush()

        self.assertEqual(["Hey!"], result)

    def test_with_failing_deferred(self):
        """
        If the target object method returns a failing L{Deferred}, a
        L{MethodCallError} is raised.
        """
        self.object.deferred = Deferred()
        self.object.method = lambda: self.object.deferred
        result = []
        deferred = self.sender.send_method_call(method="method",
                                                args=[],
                                                kwargs={})
        deferred.addErrback(result.append)

        self.connection.flush()

        # At this point the receiver is waiting for method to complete, so
        # the deferred has not fired yet
        self.assertEqual([], result)

        # Simulate time advancing and the receiver responding
        self.object.deferred.errback(Exception())
        self.connection.flush()

        [failure] = result
        failure.trap(MethodCallError)

    def test_with_deferred_timeout(self):
        """
        If the peer protocol doesn't send a response for a deferred within
        the given timeout, the method call fails.
        """
        self.object.method = lambda: Deferred()
        result = []
        deferred = self.sender.send_method_call(method="method",
                                                args=[],
                                                kwargs={})
        deferred.addErrback(result.append)

        self.clock.advance(60)

        [failure] = result
        failure.trap(MethodCallError)

    def test_with_late_response(self):
        """
        If the peer protocol sends a late response for a request that has
        already timeout, that response is ignored.
        """
        self.object.deferred = Deferred()
        self.object.method = lambda: self.object.deferred
        result = []
        deferred = self.sender.send_method_call(method="method",
                                                args=[],
                                                kwargs={})
        deferred.addErrback(result.append)

        self.clock.advance(60)
        self.object.deferred.callback("late")

        [failure] = result
        failure.trap(MethodCallError)


class RemoteObjectTest(LandscapeTest):

    def setUp(self):
        super(RemoteObjectTest, self).setUp()
        self.methods = ["method"]
        self.object = DummyObject()
        self.clock = Clock()
        self.factory = MethodCallClientFactory(self.clock)
        server_factory = MethodCallServerFactory(self.object, self.methods)
        self.connector = FakeConnector(self.factory, server_factory)
        self.connector.connect()
        self.remote = self.successResultOf(self.factory.getRemoteObject())

    def test_with_forbidden_method(self):
        """
        A L{RemoteObject} can send L{MethodCall}s without arguments and withj
        an empty response.
        """
        self.methods.remove("method")
        deferred = self.remote.method()
        failure = self.failureResultOf(deferred)
        self.assertEqual("Forbidden method 'method'", str(failure.value))

    def test_with_no_arguments(self):
        """
        A L{RemoteObject} can send L{MethodCall}s without arguments and with
        an empty response.
        """
        self.object.method = lambda: None
        deferred = self.remote.method()
        self.assertIs(None, self.successResultOf(deferred))

    def test_with_return_value(self):
        """
        A L{RemoteObject} can send L{MethodCall}s without arguments and get
        back the value of the commands's response.
        """
        self.object.method = lambda: "Cool"
        deferred = self.remote.method()
        self.assertEqual("Cool", self.successResultOf(deferred))

    def test_with_arguments(self):
        """
        A L{RemoteObject} can send L{MethodCall}s with one argument and get
        the response value.
        """
        self.object.method = lambda word, times=2: word * times
        deferred = self.remote.method("hi", times=3)
        self.assertEqual("hihihi", self.successResultOf(deferred))

    def test_method_call_error(self):
        """
        If a L{MethodCall} fails due to a L{MethodCallError},
        the L{RemoteObject} won't try to perform it again, even if the
        C{retryOnReconnect} error is set, as a L{MethodCallError} is a
        permanent failure that is not likely to ever succeed.
        """
        self.methods.remove("method")
        self.factory.retryOnReconnect = True
        deferred = self.remote.method()
        self.failureResultOf(deferred).trap(MethodCallError)

    def test_retry(self):
        """
        If the connection is lost and C{retryOnReconnect} is C{True} on the
        factory, the L{RemoteObject} will transparently retry to perform
        the L{MethodCall} requests that failed due to the broken connections.
        """
        self.object.method = lambda word: word.capitalize()
        self.factory.factor = 0.19
        self.factory.retryOnReconnect = True
        self.connector.disconnect()
        deferred = self.remote.method("john")

        # The deferred has not fired yet, because it's been put in the pending
        # queue, until the call gets a chance to be retried upon reconnection
        self.assertFalse(deferred.called)

        # Time passes and the factory successfully reconnects
        self.clock.advance(1)

        # We finally get the result
        self.assertEqual("John", self.successResultOf(deferred))

    def test_retry_with_method_call_error(self):
        """
        If a retried L{MethodCall} request fails due to a L{MethodCallError},
        the L{RemoteObject} will properly propagate the error to the original
        caller.
        """
        self.methods.remove("method")
        self.factory.factor = 0.19
        self.factory.retryOnReconnect = True
        self.connector.disconnect()
        deferred = self.remote.method()

        # The deferred has not fired yet, because it's been put in the pending
        # queue, until the call gets a chance to be retried upon reconnection
        self.assertFalse(deferred.called)

        # Time passes and the factory successfully reconnects
        self.clock.advance(1)

        failure = self.failureResultOf(deferred)
        self.assertEqual("Forbidden method 'method'", str(failure.value))


class MethodCallClientFactoryTest(LandscapeTest):

    def setUp(self):
        super(MethodCallClientFactoryTest, self).setUp()
        self.clock = Clock()
        self.factory = MethodCallClientFactory(self.clock)

    def test_max_delay(self):
        """
        The L{MethodCallClientFactory} class has a default value of 30 seconds
        for the maximum reconnection delay.
        """
        self.assertEqual(self.factory.maxDelay, 30)

    def test_connect_notifier(self):
        """
        The C{notifyOnConnect} method supports specifying a callback that
        will be invoked when a connection has been established.
        """
        protocols = []
        self.factory.notifyOnConnect(protocols.append)
        protocol = self.factory.buildProtocol(None)
        protocol.connectionMade()
        self.assertEqual([protocol], protocols)

    def test_connect_notifier_with_reconnect(self):
        """
        The C{notifyOnConnect} method will also callback when a connection is
        re-established after it was lost.
        """
        protocols = []
        self.factory.notifyOnConnect(protocols.append)
        protocol1 = self.factory.buildProtocol(None)
        protocol1.connectionMade()
        protocol2 = self.factory.buildProtocol(None)
        protocol2.connectionMade()
        self.assertEqual([protocol1, protocol2], protocols)

    def test_get_remote_object(self):
        """
        The C{getRemoteObject} method returns a deferred firing with a
        connected L{RemoteBroker}.
        """
        deferred = self.factory.getRemoteObject()
        protocol = self.factory.buildProtocol(None)
        protocol.connectionMade()
        self.assertIsInstance(self.successResultOf(deferred), RemoteObject)

    def test_get_remote_object_failure(self):
        """
        If the factory fails to establish a connection the deferreds returned
        by C{getRemoteObject} will fail.
        """
        deferred = self.factory.getRemoteObject()
        self.factory.continueTrying = False  # Don't retry
        self.factory.clientConnectionFailed(None, Failure(ConnectError()))
        self.failureResultOf(deferred).trap(ConnectError)

    def test_client_connection_failed(self):
        """
        The L{MethodCallClientFactory} keeps trying to connect if maxRetries
        is not reached.
        """
        class FakeConnector(object):
            called = False

            def connect(self):
                self.called = True

        connector = FakeConnector()
        self.assertEqual(self.factory.retries, 0)
        self.factory.clientConnectionFailed(connector, None)
        self.assertEqual(self.factory.retries, 1)
        self.clock.advance(5)
        self.assertTrue(connector.called)

    def test_reconnect(self):
        """
        If the connection is lost, the L{RemoteObject} created by the creator
        will transparently handle the reconnection.
        """
        dummy_object = DummyObject()
        dummy_object.method = lambda: None
        server_factory = MethodCallServerFactory(dummy_object, ["method"])
        connector = FakeConnector(self.factory, server_factory)
        connector.connect()
        remote = self.successResultOf(self.factory.getRemoteObject())

        connector.disconnect()
        self.clock.advance(5)
        deferred = remote.method()
        self.assertIs(None, self.successResultOf(deferred))


class MethodCallFunctionalTest(LandscapeTest):

    def setUp(self):
        super(MethodCallFunctionalTest, self).setUp()
        self.methods = ["method"]
        self.object = DummyObject()
        self.object.method = lambda word: word.capitalize()
        self.socket = self.mktemp()
        self.server = MethodCallServerFactory(self.object, self.methods)
        self.client = MethodCallClientFactory(reactor)
        self.port = reactor.listenUNIX(self.socket, self.server)

    def tearDown(self):
        super(MethodCallFunctionalTest, self).tearDown()
        self.port.stopListening()

    @inlineCallbacks
    def test_connect(self):
        """
        The L{RemoteObject} resulting form the deferred returned by
        L{MethodCallClientFactory.getRemoteObject} is properly connected
        to the remote peer.
        """
        connector = reactor.connectUNIX(self.socket, self.client)
        remote = yield self.client.getRemoteObject()
        result = yield remote.method("john")
        self.assertEqual(result, "John")
        self.client.stopTrying()
        connector.disconnect()

    @inlineCallbacks
    def test_connect_with_max_retries(self):
        """
        If L{MethodCallClientFactory.maxRetries} is set, then the factory
        will give up trying to connect after that amout of times.
        """
        self.port.stopListening()
        self.client.maxRetries = 0
        reactor.connectUNIX(self.socket, self.client)
        yield self.assertFailure(self.client.getRemoteObject(), ConnectError)

    @inlineCallbacks
    def test_reconnect(self):
        """
        If the connection is lost, the L{RemoteObject} created by the factory
        will transparently handle the reconnection.
        """
        self.client.factor = 0.01  # Try reconnecting very quickly
        connector = reactor.connectUNIX(self.socket, self.client)
        remote = yield self.client.getRemoteObject()

        # Disconnect and wait till we connect again
        deferred = Deferred()
        self.client.notifyOnConnect(deferred.callback)
        connector.disconnect()
        yield deferred

        # The remote object is still working
        result = yield remote.method("john")
        self.assertEqual(result, "John")
        self.client.stopTrying()
        connector.disconnect()

    @inlineCallbacks
    def test_retry(self):
        """
        If the connection is lost, the L{RemoteObject} created by the creator
        will transparently retry to perform the L{MethodCall} requests that
        failed due to the broken connection.
        """
        self.client.factor = 0.01  # Try reconnecting very quickly
        self.client.retryOnReconnect = True
        connector = reactor.connectUNIX(self.socket, self.client)
        remote = yield self.client.getRemoteObject()

        # Disconnect
        connector.disconnect()

        # This call will fail but it's transparently retried
        result = yield remote.method("john")
        self.assertEqual(result, "John")
        self.client.stopTrying()
        connector.disconnect()

    @inlineCallbacks
    def test_retry_with_method_call_error(self):
        """
        If a retried L{MethodCall} request fails due to a L{MethodCallError},
        the L{RemoteObject} will properly propagate the error to the original
        caller.
        """
        self.methods.remove("method")
        self.client.factor = 0.01  # Try reconnecting very quickly
        self.client.retryOnReconnect = True
        connector = reactor.connectUNIX(self.socket, self.client)
        remote = yield self.client.getRemoteObject()

        # Disconnect
        connector.disconnect()

        # A method call error is not retried
        yield self.assertFailure(remote.method(), MethodCallError)
        self.client.stopTrying()
        connector.disconnect()

    @inlineCallbacks
    def test_wb_retry_with_while_still_disconnected(self):
        """
        The L{RemoteObject._retry} method gets called as soon as a new
        connection is ready. If for whatever reason the connection drops
        again very quickly, the C{_retry} method will behave as expected.
        """
        self.methods.remove("method")
        self.client.factor = 0.01  # Try reconnecting very quickly
        self.client.retryOnReconnect = True
        connector = reactor.connectUNIX(self.socket, self.client)
        remote = yield self.client.getRemoteObject()

        # Disconnect
        connector.disconnect()

        def handle_reconnect(protocol):
            # In this precise moment we have a newly connected protocol
            remote._sender._protocol = protocol

            # Pretend that the connection is lost again very quickly
            protocol.transport.loseConnection()

            # Force RemoteObject._retry to run using a disconnected protocol
            reactor.callLater(0, remote._retry)

            # Restore the real handler and start listening again very soon
            self.client.dontNotifyOnConnect(handle_reconnect)
            self.client.notifyOnConnect(remote._handle_connect)

        def assert_failure(error):
            self.assertEqual(str(error), "Forbidden method 'method'")

        # Use our own reconnect handler
        self.client.dontNotifyOnConnect(remote._handle_connect)
        self.client.notifyOnConnect(handle_reconnect)

        error = yield self.assertFailure(remote.method(), MethodCallError)
        self.assertEqual(str(error), "Forbidden method 'method'")

        self.client.stopTrying()
        connector.disconnect()

    @inlineCallbacks
    def test_retry_with_many_method_calls(self):
        """
        If several L{MethodCall} requests were issued while disconnected, they
        will be all eventually completed when the connection gets established
        again.
        """
        self.client.factor = 0.01  # Try reconnecting very quickly
        self.client.retryOnReconnect = True
        connector = reactor.connectUNIX(self.socket, self.client)
        remote = yield self.client.getRemoteObject()

        # Disconnect
        connector.disconnect()

        result1 = yield remote.method("john")
        result2 = yield remote.method("bill")

        self.assertEqual(result1, "John")
        self.assertEqual(result2, "Bill")
        self.client.stopTrying()
        connector.disconnect()

    @inlineCallbacks
    def test_retry_without_retry_on_reconnect(self):
        """
        If C{retryOnReconnect} is C{False}, the L{RemoteObject} object won't
        retry to perform requests which failed because the connection was
        lost, however requests made after a reconnection will still succeed.
        """
        self.client.factor = 0.01  # Try reconnecting very quickly
        connector = reactor.connectUNIX(self.socket, self.client)
        remote = yield self.client.getRemoteObject()

        # Disconnect
        deferred = Deferred()
        self.client.notifyOnConnect(deferred.callback)
        connector.disconnect()

        yield self.assertFailure(remote.modt(), ConnectionDone)

        # Wait for reconnection and peform another call
        yield deferred
        result = yield remote.method("john")
        self.assertEqual(result, "John")

        self.client.stopTrying()
        connector.disconnect()

    @inlineCallbacks
    def test_retry_with_timeout(self):
        """
        If a C{retryTimeout} is set, the L{RemoteObject} object will errback
        failed L{MethodCall}s after that amount of seconds, without retrying
        them when the connection established again.
        """
        self.client.retryOnReconnect = True
        self.client.retryTimeout = 0.1
        self.client.factor = 1  # Reconnect slower than timeout
        connector = reactor.connectUNIX(self.socket, self.client)
        remote = yield self.client.getRemoteObject()

        # Disconnect
        connector.disconnect()

        error = yield self.assertFailure(remote.method("foo"), MethodCallError)
        self.assertEqual("timeout", str(error))

        self.client.stopTrying()
        connector.disconnect()



import os

import pycurl

from twisted.internet.defer import FirstError

from landscape.lib.fetch import (
    fetch, fetch_async, fetch_many_async, fetch_to_files,
    url_to_filename, HTTPCodeError, PyCurlError)
from landscape.tests.helpers import LandscapeTest


class CurlStub(object):

    def __init__(self, result=None, infos=None, error=None):
        self.result = result
        self.infos = infos
        if self.infos is None:
            self.infos = {pycurl.HTTP_CODE: 200}
        self.options = {}
        self.performed = False
        self.error = error

    def getinfo(self, what):
        if what in self.infos:
            return self.infos[what]
        raise RuntimeError("Stub doesn't know about %d info" % what)

    def setopt(self, option, value):
        if isinstance(value, unicode):
            raise AssertionError("setopt() doesn't accept unicode values")
        if self.performed:
            raise AssertionError("setopt() can't be called after perform()")
        self.options[option] = value

    def perform(self):
        if self.error:
            raise self.error
        if self.performed:
            raise AssertionError("Can't perform twice")
        self.options[pycurl.WRITEFUNCTION](self.result)
        self.performed = True


class CurlManyStub(object):

    def __init__(self, url_results):
        self.curls = {}
        for url in url_results:
            result = url_results[url]
            if isinstance(result, str):
                body = result
                http_code = 200
            else:
                body = result[0]
                http_code = result[1]
            self.curls[url] = CurlStub(body, {pycurl.HTTP_CODE: http_code})
        self.current = None

    def getinfo(self, what):
        if not self.current.performed:
            raise AssertionError("getinfo() can't be called before perform()")
        result = self.current.getinfo(what)
        self.current = None
        return result

    def setopt(self, option, value):
        if option is pycurl.URL:
            self.current = self.curls[value]
        self.current.setopt(option, value)

    def perform(self):
        self.current.perform()


class Any(object):

    def __eq__(self, other):
        return True


class FetchTest(LandscapeTest):

    def test_basic(self):
        curl = CurlStub("result")
        result = fetch("http://example.com", curl=curl)
        self.assertEqual(result, "result")
        self.assertEqual(curl.options,
                         {pycurl.URL: "http://example.com",
                          pycurl.FOLLOWLOCATION: 1,
                          pycurl.MAXREDIRS: 5,
                          pycurl.CONNECTTIMEOUT: 30,
                          pycurl.LOW_SPEED_LIMIT: 1,
                          pycurl.LOW_SPEED_TIME: 600,
                          pycurl.NOSIGNAL: 1,
                          pycurl.WRITEFUNCTION: Any(),
                          pycurl.DNS_CACHE_TIMEOUT: 0,
                          pycurl.ENCODING: "gzip,deflate"})

    def test_post(self):
        curl = CurlStub("result")
        result = fetch("http://example.com", post=True, curl=curl)
        self.assertEqual(result, "result")
        self.assertEqual(curl.options,
                         {pycurl.URL: "http://example.com",
                          pycurl.FOLLOWLOCATION: 1,
                          pycurl.MAXREDIRS: 5,
                          pycurl.CONNECTTIMEOUT: 30,
                          pycurl.LOW_SPEED_LIMIT: 1,
                          pycurl.LOW_SPEED_TIME: 600,
                          pycurl.NOSIGNAL: 1,
                          pycurl.WRITEFUNCTION: Any(),
                          pycurl.POST: True,
                          pycurl.DNS_CACHE_TIMEOUT: 0,
                          pycurl.ENCODING: "gzip,deflate"})

    def test_post_data(self):
        curl = CurlStub("result")
        result = fetch("http://example.com", post=True, data="data", curl=curl)
        self.assertEqual(result, "result")
        self.assertEqual(curl.options[pycurl.READFUNCTION](), "data")
        self.assertEqual(curl.options,
                         {pycurl.URL: "http://example.com",
                          pycurl.FOLLOWLOCATION: 1,
                          pycurl.MAXREDIRS: 5,
                          pycurl.CONNECTTIMEOUT: 30,
                          pycurl.LOW_SPEED_LIMIT: 1,
                          pycurl.LOW_SPEED_TIME: 600,
                          pycurl.NOSIGNAL: 1,
                          pycurl.WRITEFUNCTION: Any(),
                          pycurl.POST: True,
                          pycurl.POSTFIELDSIZE: 4,
                          pycurl.READFUNCTION: Any(),
                          pycurl.DNS_CACHE_TIMEOUT: 0,
                          pycurl.ENCODING: "gzip,deflate"})

    def test_cainfo(self):
        curl = CurlStub("result")
        result = fetch("https://example.com", cainfo="cainfo", curl=curl)
        self.assertEqual(result, "result")
        self.assertEqual(curl.options,
                         {pycurl.URL: "https://example.com",
                          pycurl.FOLLOWLOCATION: 1,
                          pycurl.MAXREDIRS: 5,
                          pycurl.CONNECTTIMEOUT: 30,
                          pycurl.LOW_SPEED_LIMIT: 1,
                          pycurl.LOW_SPEED_TIME: 600,
                          pycurl.NOSIGNAL: 1,
                          pycurl.WRITEFUNCTION: Any(),
                          pycurl.CAINFO: "cainfo",
                          pycurl.DNS_CACHE_TIMEOUT: 0,
                          pycurl.ENCODING: "gzip,deflate"})

    def test_cainfo_on_http(self):
        curl = CurlStub("result")
        result = fetch("http://example.com", cainfo="cainfo", curl=curl)
        self.assertEqual(result, "result")
        self.assertTrue(pycurl.CAINFO not in curl.options)

    def test_headers(self):
        curl = CurlStub("result")
        result = fetch("http://example.com",
                       headers={"a": "1", "b": "2"}, curl=curl)
        self.assertEqual(result, "result")
        self.assertEqual(curl.options,
                         {pycurl.URL: "http://example.com",
                          pycurl.FOLLOWLOCATION: 1,
                          pycurl.MAXREDIRS: 5,
                          pycurl.CONNECTTIMEOUT: 30,
                          pycurl.LOW_SPEED_LIMIT: 1,
                          pycurl.LOW_SPEED_TIME: 600,
                          pycurl.NOSIGNAL: 1,
                          pycurl.WRITEFUNCTION: Any(),
                          pycurl.HTTPHEADER: ["a: 1", "b: 2"],
                          pycurl.DNS_CACHE_TIMEOUT: 0,
                          pycurl.ENCODING: "gzip,deflate"})

    def test_timeouts(self):
        curl = CurlStub("result")
        result = fetch("http://example.com", connect_timeout=5,
                       total_timeout=30, curl=curl)
        self.assertEqual(result, "result")
        self.assertEqual(curl.options,
                         {pycurl.URL: "http://example.com",
                          pycurl.FOLLOWLOCATION: 1,
                          pycurl.MAXREDIRS: 5,
                          pycurl.CONNECTTIMEOUT: 5,
                          pycurl.LOW_SPEED_LIMIT: 1,
                          pycurl.LOW_SPEED_TIME: 30,
                          pycurl.NOSIGNAL: 1,
                          pycurl.WRITEFUNCTION: Any(),
                          pycurl.DNS_CACHE_TIMEOUT: 0,
                          pycurl.ENCODING: "gzip,deflate"})

    def test_unicode(self):
        """
        The L{fetch} function converts the C{url} parameter to C{str} before
        passing it to curl.
        """
        curl = CurlStub("result")
        result = fetch(u"http://example.com", curl=curl)
        self.assertEqual(result, "result")
        self.assertEqual(curl.options[pycurl.URL], "http://example.com")
        self.assertTrue(isinstance(curl.options[pycurl.URL], str))

    def test_non_200_result(self):
        curl = CurlStub("result", {pycurl.HTTP_CODE: 404})
        try:
            fetch("http://example.com", curl=curl)
        except HTTPCodeError, error:
            self.assertEqual(error.http_code, 404)
            self.assertEqual(error.body, "result")
        else:
            self.fail("HTTPCodeError not raised")

    def test_http_error_str(self):
        self.assertEqual(str(HTTPCodeError(501, "")),
                         "Server returned HTTP code 501")

    def test_http_error_repr(self):
        self.assertEqual(repr(HTTPCodeError(501, "")),
                         "<HTTPCodeError http_code=501>")

    def test_pycurl_error(self):
        curl = CurlStub(error=pycurl.error(60, "pycurl error"))
        try:
            fetch("http://example.com", curl=curl)
        except PyCurlError, error:
            self.assertEqual(error.error_code, 60)
            self.assertEqual(error.message, "pycurl error")
        else:
            self.fail("PyCurlError not raised")

    def test_pycurl_insecure(self):
        curl = CurlStub("result")
        result = fetch("http://example.com/get-ca-cert", curl=curl,
                       insecure=True)
        self.assertEqual(result, "result")
        self.assertEqual(curl.options,
                         {pycurl.URL: "http://example.com/get-ca-cert",
                          pycurl.FOLLOWLOCATION: 1,
                          pycurl.MAXREDIRS: 5,
                          pycurl.CONNECTTIMEOUT: 30,
                          pycurl.LOW_SPEED_LIMIT: 1,
                          pycurl.LOW_SPEED_TIME: 600,
                          pycurl.NOSIGNAL: 1,
                          pycurl.WRITEFUNCTION: Any(),
                          pycurl.SSL_VERIFYPEER: False,
                          pycurl.DNS_CACHE_TIMEOUT: 0,
                          pycurl.ENCODING: "gzip,deflate"})

    def test_pycurl_error_str(self):
        self.assertEqual(str(PyCurlError(60, "pycurl error")),
                         "Error 60: pycurl error")

    def test_pycurl_error_repr(self):
        self.assertEqual(repr(PyCurlError(60, "pycurl error")),
                         "<PyCurlError args=(60, 'pycurl error')>")

    def test_pycurl_follow_true(self):
        curl = CurlStub("result")
        result = fetch("http://example.com", curl=curl,
                       follow=True)
        self.assertEqual(result, "result")
        self.assertEqual(1, curl.options[pycurl.FOLLOWLOCATION])

    def test_pycurl_follow_false(self):
        curl = CurlStub("result")
        result = fetch("http://example.com", curl=curl,
                       follow=False)
        self.assertEqual(result, "result")
        self.assertNotIn(pycurl.FOLLOWLOCATION, curl.options.keys())

    def test_create_curl(self):
        curls = []

        def pycurl_Curl():
            curl = CurlStub("result")
            curls.append(curl)
            return curl
        Curl = pycurl.Curl
        try:
            pycurl.Curl = pycurl_Curl
            result = fetch("http://example.com")
            curl = curls[0]
            self.assertEqual(result, "result")
            self.assertEqual(curl.options,
                             {pycurl.URL: "http://example.com",
                              pycurl.FOLLOWLOCATION: 1,
                              pycurl.MAXREDIRS: 5,
                              pycurl.CONNECTTIMEOUT: 30,
                              pycurl.LOW_SPEED_LIMIT: 1,
                              pycurl.LOW_SPEED_TIME: 600,
                              pycurl.NOSIGNAL: 1,
                              pycurl.WRITEFUNCTION: Any(),
                              pycurl.DNS_CACHE_TIMEOUT: 0,
                              pycurl.ENCODING: "gzip,deflate"})
        finally:
            pycurl.Curl = Curl

    def test_async_fetch(self):
        curl = CurlStub("result")
        d = fetch_async("http://example.com/", curl=curl)

        def got_result(result):
            self.assertEqual(result, "result")
        return d.addCallback(got_result)

    def test_async_fetch_with_error(self):
        curl = CurlStub("result", {pycurl.HTTP_CODE: 501})
        d = fetch_async("http://example.com/", curl=curl)

        def got_error(failure):
            self.assertEqual(failure.value.http_code, 501)
            self.assertEqual(failure.value.body, "result")
            return failure
        d.addErrback(got_error)
        self.assertFailure(d, HTTPCodeError)
        return d

    def test_fetch_many_async(self):
        """
        L{fetch_many_async} retrieves multiple URLs, and returns a
        C{DeferredList} firing its callback when all the URLs have
        successfully completed.
        """
        url_results = {"http://good/": "good",
                       "http://better/": "better"}

        def callback(result, url):
            self.assertIn(result, url_results.values())
            self.assertIn(url, url_results)
            url_results.pop(url)

        def errback(failure, url):
            self.fail()

        curl = CurlManyStub(url_results)
        d = fetch_many_async(url_results.keys(), callback=callback,
                             errback=errback, curl=curl)

        def completed(result):
            self.assertEqual(url_results, {})

        return d.addCallback(completed)

    def test_fetch_many_async_with_error(self):
        """
        L{fetch_many_async} aborts as soon as one URL fails.
        """
        url_results = {"http://right/": "right",
                       "http://wrong/": ("wrong", 501),
                       "http://impossible/": "impossible"}
        failed_urls = []

        def errback(failure, url):
            failed_urls.append(url)
            self.assertEqual(failure.value.body, "wrong")
            self.assertEqual(failure.value.http_code, 501)
            return failure

        curl = CurlManyStub(url_results)
        urls = ["http://right/", "http://wrong/", "http://impossible/"]
        result = fetch_many_async(urls, callback=None,
                                  errback=errback, curl=curl)

        def check_failure(failure):
            self.assertTrue(isinstance(failure.subFailure.value,
                                       HTTPCodeError))
            self.assertEqual(failed_urls, ["http://wrong/"])

        self.assertFailure(result, FirstError)
        return result.addCallback(check_failure)

    def test_url_to_filename(self):
        """
        L{url_to_filename} extracts the filename part of an URL, optionally
        prepending a directory path to it.
        """
        self.assertEqual(url_to_filename("http://some/file"), "file")
        self.assertEqual(url_to_filename("http://some/file/"), "file")
        self.assertEqual(url_to_filename("http://some/file", directory="dir"),
                         os.path.join("dir", "file"))

    def test_fetch_to_files(self):
        """
        L{fetch_to_files} fetches a list of URLs and save their content
        in the given directory.
        """
        url_results = {"http://good/file": "file",
                       "http://even/better-file": "better-file"}
        directory = self.makeDir()
        curl = CurlManyStub(url_results)

        result = fetch_to_files(url_results.keys(), directory, curl=curl)

        def check_files(ignored):
            for result in url_results.itervalues():
                fd = open(os.path.join(directory, result))
                self.assertEqual(fd.read(), result)
                fd.close()

        result.addCallback(check_files)
        return result

    def test_fetch_to_files_with_trailing_slash(self):
        """
        L{fetch_to_files} discards trailing slashes from the final component
        of the given URLs when saving them as files.
        """
        directory = self.makeDir()
        curl = CurlStub("data")

        result = fetch_to_files(["http:///with/slash/"], directory, curl=curl)

        def check_files(ignored):
            os.path.exists(os.path.join(directory, "slash"))

        result.addCallback(check_files)
        return result

    def test_fetch_to_files_with_errors(self):
        """
        L{fetch_to_files} optionally logs an error message as soon as one URL
        fails, and aborts.
        """
        url_results = {"http://im/right": "right",
                       "http://im/wrong": ("wrong", 404),
                       "http://im/not": "not"}
        directory = self.makeDir()
        messages = []
        logger = lambda message: messages.append(message)
        curl = CurlManyStub(url_results)

        result = fetch_to_files(url_results.keys(), directory, logger=logger,
                                curl=curl)

        def check_messages(failure):
            self.assertEqual(len(messages), 1)
            self.assertEqual(messages[0],
                             "Couldn't fetch file from http://im/wrong "
                             "(Server returned HTTP code 404)")
            messages.pop()

        def check_files(ignored):
            self.assertEqual(messages, [])
            self.assertFalse(os.path.exists(os.path.join(directory, "wrong")))

        result.addErrback(check_messages)
        result.addCallback(check_files)
        return result

    def test_fetch_to_files_with_non_existing_directory(self):
        """
        The deferred list returned by L{fetch_to_files} results in a failure
        if the destination directory doesn't exist.
        """
        url_results = {"http://im/right": "right"}
        directory = "i/dont/exist/"
        curl = CurlManyStub(url_results)

        result = fetch_to_files(url_results.keys(), directory, curl=curl)

        def check_error(failure):
            error = str(failure.value.subFailure.value)
            self.assertEqual(error, "[Errno 2] No such file or directory: "
                              "'i/dont/exist/right'")
            self.assertFalse(os.path.exists(os.path.join(directory, "right")))

        result.addErrback(check_error)
        return result



import sys

import dbus
import dbus.service

from landscape.ui.lib.polkit import PolicyKitMechanism, POLICY_NAME


SERVICE_NAME = "com.canonical.LandscapeClientSettings"
INTERFACE_NAME = "com.canonical.LandscapeClientSettings.ConfigurationInterface"
OBJECT_PATH = "/com/canonical/LandscapeClientSettings/ConfigurationInterface"


class PermissionDeniedByPolicy(dbus.DBusException):
    _dbus_error_name = \
        "com.canonical.LandscapeClientSettings.PermissionDeniedByPolicy"


class ConfigurationMechanism(PolicyKitMechanism):
    """
    L{ConfigurationMechanism} provides access to the
    L{LandscapeSetupConfiguration} object via DBus with access control
    implemented via PolicyKit policy.  The use of DBus results from the use of
    PolicyKit, not the other way around, and is done that way because that is
    considered to be the Right Thing for Ubuntu Desktop circa January 2012.
    """

    def __init__(self, config, bus_name, bypass=False, conn=None):
        super(ConfigurationMechanism, self).__init__(
            OBJECT_PATH, bus_name, PermissionDeniedByPolicy,
            bypass=bypass, conn=conn)
        self._config = config

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="as",
                         out_signature="",
                         sender_keyword="sender",
                         connection_keyword="conn")
    def load(self, arglist, sender=None, conn=None):
        if self._is_allowed_by_policy(sender, conn, POLICY_NAME):
            if len(arglist) > 0:
                self._config.load(arglist.split(chr(0x1e)))
            else:
                self._config.load([])
        return

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="", out_signature="",
                         sender_keyword="sender", connection_keyword="conn")
    def reload(self, sender=None, conn=None):
        if self._is_allowed_by_policy(sender, conn, POLICY_NAME):
            self._config.reload()

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="", out_signature="",
                         sender_keyword="sender", connection_keyword="conn")
    def write(self, sender=None, conn=None):
        if self._is_allowed_by_policy(sender, conn, POLICY_NAME):
            self._config.write()

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="",
                         out_signature="s",
                         sender_keyword="sender",
                         connection_keyword="conn")
    def get_config_filename(self, sender=None, conn=None):
        return self._config.get_config_filename()

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="s",
                         out_signature="s",
                         sender_keyword="sender",
                         connection_keyword="conn")
    def get(self, name, sender=None, conn=None):
        """
        Return the configuration option value associated with L{name} from the
        L{LandscapeSetupConfiguration}.
        """
        if self._is_allowed_by_policy(sender, conn, POLICY_NAME):
            try:
                value = self._config.get(name)
            except AttributeError:
                return ""
            if value is None:
                return ""
            return str(value)
        return ""

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="ss",
                         out_signature="",
                         sender_keyword="sender",
                         connection_keyword="conn")
    def set(self, name, value, sender=None, conn=None):
        """
        Set the configuration option associated with L{name} to L{value} in the
        L{LandscapeSetupConfiguration}.
        """
        if self._is_allowed_by_policy(sender, conn, POLICY_NAME):
            # Underlying _config does not support unicode so convert to ascii
            value = unicode(value).encode("ascii", errors="replace")
            setattr(self._config, name, value)

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="",
                         out_signature="",
                         sender_keyword="sender",
                         connection_keyword="conn")
    def exit(self, sender=None, conn=None):
        """
        Exit this process.
        """
        sys.exit(0)



import logging

from twisted.internet.defer import Deferred
from twisted.internet.protocol import ProcessProtocol
from twisted.internet.error import ProcessDone

from landscape.manager.plugin import ManagerPlugin, SUCCEEDED, FAILED


class ShutdownFailedError(Exception):
    """Raised when a call to C{/sbin/shutdown} fails.

    @ivar data: The data that the process printed before failing.
    """

    def __init__(self, data):
        self.data = data


class ShutdownManager(ManagerPlugin):

    def __init__(self, process_factory=None):
        if process_factory is None:
            from twisted.internet import reactor as process_factory
        self._process_factory = process_factory

    def register(self, registry):
        """Add this plugin to C{registry}.

        The shutdown manager handles C{shutdown} activity messages broadcast
        from the server.
        """
        super(ShutdownManager, self).register(registry)
        registry.register_message("shutdown", self.perform_shutdown)

    def perform_shutdown(self, message):
        """Request a system restart or shutdown.

        If the call to C{/sbin/shutdown} runs without errors the activity
        specified in the message will be responded as succeeded.  Otherwise,
        it will be responded as failed.
        """
        operation_id = message["operation-id"]
        reboot = message["reboot"]
        protocol = ShutdownProcessProtocol()
        protocol.set_timeout(self.registry.reactor)
        protocol.result.addCallback(self._respond_success, operation_id)
        protocol.result.addErrback(self._respond_failure, operation_id)
        command, args = self._get_command_and_args(protocol, reboot)
        self._process_factory.spawnProcess(protocol, command, args=args)

    def _respond_success(self, data, operation_id):
        logging.info("Shutdown request succeeded.")
        deferred = self._respond(SUCCEEDED, data, operation_id)
        # After sending the result to the server, stop accepting messages and
        # wait for the reboot/shutdown.
        deferred.addCallback(
            lambda _: self.registry.broker.stop_exchanger())
        return deferred

    def _respond_failure(self, failure, operation_id):
        logging.info("Shutdown request failed.")
        return self._respond(FAILED, failure.value.data, operation_id)

    def _respond(self, status, data, operation_id):
        message = {"type": "operation-result",
                   "status": status,
                   "result-text": data,
                   "operation-id": operation_id}
        return self.registry.broker.send_message(
            message, self._session_id, True)

    def _get_command_and_args(self, protocol, reboot):
        """
        Returns a C{command, args} 2-tuple suitable for use with
        L{IReactorProcess.spawnProcess}.
        """
        minutes = "+%d" % (protocol.delay // 60,)
        if reboot:
            args = ["/sbin/shutdown", "-r", minutes,
                    "Landscape is rebooting the system"]
        else:
            args = ["/sbin/shutdown", "-h", minutes,
                    "Landscape is shutting down the system"]
        return "/sbin/shutdown", args


class ShutdownProcessProtocol(ProcessProtocol):
    """A ProcessProtocol for calling C{/sbin/shutdown}.

    C{shutdown} doesn't return immediately when a time specification is
    provided.  Failures are reported immediately after it starts and return a
    non-zero exit code.  The process protocol calls C{shutdown} and waits for
    failures for C{timeout} seconds.  If no failures are reported it fires
    C{result}'s callback with whatever output was received from the process.
    If failures are reported C{result}'s errback is fired.

    @ivar result: A L{Deferred} fired when C{shutdown} fails or
        succeeds.
    @ivar reboot: A flag indicating whether a shutdown or reboot should be
        performed.  Default is C{False}.
    @ivar delay: The time in seconds from now to schedule the shutdown.
        Default is 240 seconds.  The time will be converted to minutes using
        integer division when passed to C{shutdown}.
    """

    def __init__(self, reboot=False, delay=240):
        self.result = Deferred()
        self.reboot = reboot
        self.delay = delay
        self._data = []
        self._waiting = True

    def get_data(self):
        """Get the data printed by the subprocess."""
        return "".join(self._data)

    def set_timeout(self, reactor, timeout=10):
        """
        Set the error checking timeout, after which C{result}'s callback will
        be fired.
        """
        reactor.call_later(timeout, self._succeed)

    def childDataReceived(self, fd, data):
        """Some data was received from the child.

        Add it to our buffer to pass to C{result} when it's fired.
        """
        if self._waiting:
            self._data.append(data)

    def processEnded(self, reason):
        """Fire back the C{result} L{Deferred}.

        C{result}'s callback will be fired with the string of data received
        from the subprocess, or if the subprocess failed C{result}'s errback
        will be fired with the string of data received from the subprocess.
        """
        if self._waiting:
            if reason.check(ProcessDone):
                self._succeed()
            else:
                self.result.errback(ShutdownFailedError(self.get_data()))
                self._waiting = False

    def _succeed(self):
        """Fire C{result}'s callback with data accumulated from the process."""
        if self._waiting:
            self.result.callback(self.get_data())
            self._waiting = False



from twisted.internet.defer import maybeDeferred, execute, succeed

from landscape.lib.amp import RemoteObject, MethodCallArgument
from landscape.amp import ComponentConnector, get_remote_methods
from landscape.broker.server import BrokerServer
from landscape.broker.client import BrokerClient
from landscape.monitor.monitor import Monitor
from landscape.manager.manager import Manager


class RemoteBroker(RemoteObject):

    def call_if_accepted(self, type, callable, *args):
        """Call C{callable} if C{type} is an accepted message type."""
        deferred_types = self.get_accepted_message_types()

        def got_accepted_types(result):
            if type in result:
                return callable(*args)
        deferred_types.addCallback(got_accepted_types)
        return deferred_types

    def call_on_event(self, handlers):
        """Call a given handler as soon as a certain event occurs.

        @param handlers: A dictionary mapping event types to callables, where
            an event type is string (the name of the event). When the first of
            the given event types occurs in the broker reactor, the associated
            callable will be fired.
        """
        result = self.listen_events(handlers.keys())
        return result.addCallback(lambda event_type: handlers[event_type]())


class FakeRemoteBroker(object):
    """Looks like L{RemoteBroker}, but actually talks to local objects."""

    def __init__(self, exchanger, message_store, broker_server):
        self.exchanger = exchanger
        self.message_store = message_store
        self.broker_server = broker_server

    def __getattr__(self, name):
        """
        Pass attributes through to the real L{BrokerServer}, after checking
        that they're encodable with AMP.
        """
        original = getattr(self.broker_server, name, None)
        if (name in get_remote_methods(self.broker_server)
            and original is not None
            and callable(original)):
            def method(*args, **kwargs):
                for arg in args:
                    assert MethodCallArgument.check(arg)
                for k, v in kwargs.iteritems():
                    assert MethodCallArgument.check(v)
                return execute(original, *args, **kwargs)
            return method
        else:
            raise AttributeError(name)

    def call_if_accepted(self, type, callable, *args):
        if type in self.message_store.get_accepted_types():
            return maybeDeferred(callable, *args)
        return succeed(None)


class RemoteBrokerConnector(ComponentConnector):
    """Helper to create connections with the L{BrokerServer}."""

    remote = RemoteBroker
    component = BrokerServer


class RemoteClientConnector(ComponentConnector):
    """Helper to create connections with the L{BrokerServer}."""

    component = BrokerClient


class RemoteMonitorConnector(RemoteClientConnector):
    """Helper to create connections with the L{Monitor}."""

    component = Monitor


class RemoteManagerConnector(RemoteClientConnector):
    """Helper for creating connections with the L{Monitor}."""

    component = Manager


def get_component_registry():
    """Get a mapping of component name to connectors, for all components."""
    all_connectors = [
        RemoteBrokerConnector,
        RemoteClientConnector,
        RemoteMonitorConnector,
        RemoteManagerConnector
    ]
    return dict(
        (connector.component.name, connector)
        for connector in all_connectors)





def encode_if_needed(value):
    """
    A small helper to decode unicode to utf-8 if needed.
    """
    if isinstance(value, unicode):
        value = value.encode("utf-8")
    return value


def encode_dict_if_needed(values):
    """
    A wrapper taking a dict and passing each of the values to encode_if_needed.
    """
    for key, value in values.iteritems():
        values[key] = encode_if_needed(value)
    return values



from twisted.internet.defer import maybeDeferred

from landscape.lib.log import log_failure
from landscape.amp import ComponentPublisher, ComponentConnector, remote

from landscape.monitor.plugin import MonitorPlugin
from landscape.user.changes import UserChanges
from landscape.user.provider import UserProvider


class UserMonitor(MonitorPlugin):
    """
    A plugin which monitors the system user databases.
    """

    persist_name = "users"
    scope = "users"
    run_interval = 3600  # 1 hour
    name = "usermonitor"

    def __init__(self, provider=None):
        if provider is None:
            provider = UserProvider()
        self._provider = provider
        self._publisher = None

    def register(self, registry):
        super(UserMonitor, self).register(registry)

        self.call_on_accepted("users", self._run_detect_changes, None)

        self._publisher = ComponentPublisher(self, self.registry.reactor,
                                             self.registry.config)
        self._publisher.start()

    def stop(self):
        """Stop listening for incoming AMP connections."""
        if self._publisher:
            self._publisher.stop()
            self._publisher = None

    def _reset(self):
        """Reset user and group data."""
        super(UserMonitor, self)._reset()
        return self._run_detect_changes()

    @remote
    def detect_changes(self, operation_id=None):
        return self.registry.broker.call_if_accepted(
            "users", self._run_detect_changes, operation_id)

    run = detect_changes

    def _run_detect_changes(self, operation_id=None):
        """
        If changes are detected an C{urgent-exchange} is fired to send
        updates to the server immediately.

        @param operation_id: When present it will be included in the
            C{operation-id} field.
        """
        from landscape.manager.usermanager import RemoteUserManagerConnector
        user_manager_connector = RemoteUserManagerConnector(
            self.registry.reactor, self.registry.config)

        # We'll skip checking the locked users if we're in monitor-only mode.
        if getattr(self.registry.config, "monitor_only", False):
            result = maybeDeferred(self._detect_changes,
                                   [], operation_id)
        else:

            def get_locked_usernames(user_manager):
                return user_manager.get_locked_usernames()

            def disconnect(locked_usernames):
                user_manager_connector.disconnect()
                return locked_usernames

            result = user_manager_connector.connect()
            result.addCallback(get_locked_usernames)
            result.addCallback(disconnect)
            result.addCallback(self._detect_changes, operation_id)
            result.addErrback(lambda f: self._detect_changes([], operation_id))
        return result

    def _detect_changes(self, locked_users, operation_id=None):

        def update_snapshot(result):
            changes.snapshot()
            return result

        def log_error(result):
            log_failure(result, "Error occured calling send_message in "
                        "_detect_changes")

        self._provider.locked_users = locked_users
        changes = UserChanges(self._persist, self._provider)
        message = changes.create_diff()

        if message:
            message["type"] = "users"
            if operation_id:
                message["operation-id"] = operation_id
            result = self.registry.broker.send_message(
                message, self._session_id, urgent=True)
            result.addCallback(update_snapshot)
            result.addErrback(log_error)
            return result


class RemoteUserMonitorConnector(ComponentConnector):

    component = UserMonitor



"""Provide access to the persistent data used by the L{MessageExchange}."""
import time

try:
    import sqlite3
except ImportError:
    from pysqlite2 import dbapi2 as sqlite3

from landscape.lib.store import with_cursor


class MessageContext(object):
    """Stores a context for incoming messages that require a response.

    The context consists of

      - the "operation-id" value
      - the secure ID that was in effect when the message was received
      - the message type
      - the time when the message was received

    This data will be used to detect secure ID changes between the time at
    which the request message came in and the completion of the request.
    If the secure ID did change the result message is obolete and will not be
    sent to the server.

    @param db: the sqlite database handle.
    @param id: the database key value for this instance.
    """

    def __init__(self, db, operation_id, secure_id, message_type, timestamp):
        self._db = db
        self.operation_id = operation_id
        self.secure_id = secure_id
        self.message_type = message_type
        self.timestamp = timestamp

    @with_cursor
    def remove(self, cursor):
        cursor.execute(
            "DELETE FROM message_context WHERE operation_id=?",
            (self.operation_id,))


class ExchangeStore(object):
    """Message meta data required by the L{MessageExchange}.

    The implementation uses a SQLite database as backend, with a single table
    called "message_context", whose schema is defined in
    L{ensure_exchange_schema}.

    @param filename: The name of the file that contains the sqlite database.
    """
    _db = None

    def __init__(self, filename):
        self._filename = filename

    def _ensure_schema(self):
        ensure_exchange_schema(self._db)

    @with_cursor
    def add_message_context(
            self, cursor, operation_id, secure_id, message_type):
        """Add a L{MessageContext} with the given data."""
        params = (operation_id, secure_id, message_type, time.time())
        cursor.execute(
            "INSERT INTO message_context "
            "   (operation_id, secure_id, message_type, timestamp) "
            "   VALUES (?,?,?,?)", params)
        return MessageContext(self._db, *params)

    @with_cursor
    def get_message_context(self, cursor, operation_id):
        """The L{MessageContext} for the given C{operation_id} or C{None}."""
        cursor.execute(
            "SELECT operation_id, secure_id, message_type, timestamp "
            "FROM message_context WHERE operation_id=?", (operation_id,))
        row = cursor.fetchone()
        if row:
            return MessageContext(self._db, *row)
        else:
            return None

    @with_cursor
    def all_operation_ids(self, cursor):
        """Return all operation IDs currently stored in C{message_context}."""
        cursor.execute("SELECT operation_id FROM message_context")
        result = cursor.fetchall()
        return [row[0] for row in result]


def ensure_exchange_schema(db):
    """Create all tables needed by a L{ExchangeStore}.

    @param db: A connection to a SQLite database.
    """
    cursor = db.cursor()
    try:
        cursor.execute(
            "CREATE TABLE message_context"
            " (id INTEGER PRIMARY KEY, timestamp TIMESTAMP, "
            "  secure_id TEXT NOT NULL, operation_id INTEGER NOT NULL, "
            "  message_type text NOT NULL)")
        cursor.execute(
            "CREATE UNIQUE INDEX msgctx_operationid_idx ON "
            "message_context(operation_id)")
    except (sqlite3.OperationalError, sqlite3.DatabaseError):
        cursor.close()
        db.rollback()
    else:
        cursor.close()
        db.commit()



from landscape.tests.helpers import LandscapeTest
from landscape.patch import UpgradeManager

from landscape.upgraders import monitor


class TestMonitorUpgraders(LandscapeTest):

    def test_monitor_upgrade_manager(self):
        self.assertEqual(type(monitor.upgrade_manager), UpgradeManager)



import re


_tag_check = re.compile("^\w+[\w-]*$", re.UNICODE)


def is_valid_tag(tagname):
    """Return True if the tag meets our tag requirements."""
    return _tag_check.match(tagname)


def is_valid_tag_list(tag_list):
    """Validate a tag_list string.

    @param tag_list: string like london, server which will be split on the
    commas and each tag verified for validity.
    """
    if not tag_list:
        return True
    return all(is_valid_tag(tag.strip()) for tag in tag_list.split(","))



from logging import info

from twisted.internet.defer import succeed

from landscape.log import format_object
from landscape.lib.log import log_failure
from landscape.broker.client import BrokerClientPlugin


class MonitorPlugin(BrokerClientPlugin):
    """
    @cvar persist_name: If specified as a string, a C{_persist} attribute
    will be available after registration.
    """

    persist_name = None
    scope = None

    def register(self, monitor):
        super(MonitorPlugin, self).register(monitor)
        if self.persist_name is not None:
            self._persist = self.monitor.persist.root_at(self.persist_name)
        else:
            self._persist = None

    def _reset(self):
        if self.persist_name is not None:
            self.registry.persist.remove(self.persist_name)

    @property
    def persist(self):
        """Return our L{Persist}, if any."""
        return self._persist

    @property
    def monitor(self):
        """An alias for the C{client} attribute."""
        return self.client


class DataWatcher(MonitorPlugin):
    """
    A utility for plugins which send data to the Landscape server
    which does not constantly change. New messages will only be sent
    when the result of get_data() has changed since the last time it
    was called.

    Subclasses should provide a get_data method, and message_type,
    message_key, and persist_name class attributes.
    """

    message_type = None
    message_key = None

    def get_message(self):
        """
        Construct a message with the latest data, or None, if the data
        has not changed since the last call.
        """
        data = self.get_data()
        if self._persist.get("data") != data:
            self._persist.set("data", data)
            return {"type": self.message_type, self.message_key: data}

    def send_message(self, urgent):
        message = self.get_message()
        if message is not None:
            info("Queueing a message with updated data watcher info "
                 "for %s.", format_object(self))
            result = self.registry.broker.send_message(
                message, self._session_id, urgent=urgent)

            def persist_data(message_id):
                self.persist_data()

            result.addCallback(persist_data)
            result.addErrback(log_failure)
            return result
        return succeed(None)

    def persist_data(self):
        """
        Sub-classes that need to defer the saving of persistent data
        should override this method.
        """
        pass

    def exchange(self, urgent=False):
        """
        Conditionally add a message to the message store if new data
        is available.
        """
        return self.registry.broker.call_if_accepted(self.message_type,
                                                     self.send_message, urgent)



import os
import re
import logging

from twisted.internet.defer import succeed, Deferred

from landscape.lib.lock import lock_path, LockError
from landscape.lib.log import log_failure
from landscape.lib.lsb_release import LSB_RELEASE_FILENAME, parse_lsb_release
from landscape.reactor import LandscapeReactor
from landscape.deployment import Configuration, init_logging
from landscape.package.store import PackageStore, InvalidHashIdDb
from landscape.broker.amp import RemoteBrokerConnector


class PackageTaskError(Exception):
    """Raised when a task hasn't been successfully completed."""


class PackageTaskHandlerConfiguration(Configuration):
    """Specialized configuration for L{PackageTaskHandler}s."""

    @property
    def package_directory(self):
        """Get the path to the package directory."""
        return os.path.join(self.data_path, "package")

    @property
    def store_filename(self):
        """Get the path to the SQlite file for the L{PackageStore}."""
        return os.path.join(self.package_directory, "database")

    @property
    def hash_id_directory(self):
        """Get the path to the directory holding the stock hash-id stores."""
        return os.path.join(self.package_directory, "hash-id")

    @property
    def update_stamp_filename(self):
        """Get the path to the update-stamp file."""
        return os.path.join(self.package_directory, "update-stamp")

    @property
    def detect_package_changes_stamp(self):
        """Get the path to the stamp marking when the last time we checked for
        changes in the packages was."""
        return os.path.join(self.data_path, "detect_package_changes_timestamp")


class LazyRemoteBroker(object):
    """Wrapper class around L{RemoteBroker} providing lazy initialization.

    This class is a wrapper around a regular L{RemoteBroker}. It connects to
    the remote broker object only when one of its attributes is first accessed.

    @param connector: The L{RemoteBrokerConnector} which will be used
        to connect to the broker.

    @note: This behaviour is needed in particular by the ReleaseUpgrader and
        the PackageChanger, because if the they connect early and the
        landscape-client package gets upgraded while they run, they will lose
        the connection and will not be able to reconnect for a potentially long
        window of time (till the new landscape-client package version is fully
        configured and the service is started again).
    """

    def __init__(self, connector):
        self._connector = connector
        self._remote = None

    def __getattr__(self, method):

        if self._remote:
            return getattr(self._remote, method)

        def wrapper(*args, **kwargs):

            def got_connection(remote):
                self._remote = remote
                return getattr(self._remote, method)(*args, **kwargs)

            result = self._connector.connect()
            return result.addCallback(got_connection)

        return wrapper


class PackageTaskHandler(object):

    config_factory = PackageTaskHandlerConfiguration

    queue_name = "default"
    lsb_release_filename = LSB_RELEASE_FILENAME
    package_store_class = PackageStore

    # This file is touched after every succesful 'apt-get update' run if the
    # update-notifier-common package is installed.
    update_notifier_stamp = "/var/lib/apt/periodic/update-success-stamp"

    def __init__(self, package_store, package_facade, remote_broker, config):
        self._store = package_store
        self._facade = package_facade
        self._broker = remote_broker
        self._config = config
        self._count = 0
        self._session_id = None

    def run(self):
        return self.handle_tasks()

    def handle_tasks(self):
        """Handle the tasks in the queue.

        The tasks will be handed over one by one to L{handle_task} until the
        queue is empty or a task fails.

        @see: L{handle_tasks}
        """
        return self._handle_next_task(None)

    def _handle_next_task(self, result, last_task=None):
        """Pick the next task from the queue and pass it to C{handle_task}."""

        if last_task is not None:
            # Last task succeeded.  We can safely kill it now.
            last_task.remove()
            self._count += 1

        task = self._store.get_next_task(self.queue_name)

        if task:
            # We have another task.  Let's handle it.
            result = self.handle_task(task)
            result.addCallback(self._handle_next_task, last_task=task)
            result.addErrback(self._handle_task_failure)
            return result

        else:
            # No more tasks!  We're done!
            return succeed(None)

    def _handle_task_failure(self, failure):
        """Gracefully handle a L{PackageTaskError} and stop handling tasks."""
        failure.trap(PackageTaskError)

    def handle_task(self, task):
        """Handle a single task.

        Sub-classes must override this method in order to trigger task-specific
        actions.

        This method must return a L{Deferred} firing the task result. If the
        deferred is successful the task will be removed from the queue and the
        next one will be picked. If the task can't be completed, this method
        must raise a L{PackageTaskError}, in this case the handler will stop
        processing tasks and the failed task won't be removed from the queue.
        """
        return succeed(None)

    @property
    def handled_tasks_count(self):
        """
        Return the number of tasks that have been successfully handled so far.
        """
        return self._count

    def use_hash_id_db(self):
        """
        Attach the appropriate pre-canned hash=>id database to our store.
        """

        def use_it(hash_id_db_filename):

            if hash_id_db_filename is None:
                # Couldn't determine which hash=>id database to use,
                # just ignore the failure and go on
                return

            if not os.path.exists(hash_id_db_filename):
                # The appropriate database isn't there, but nevermind
                # and just go on
                return

            try:
                self._store.add_hash_id_db(hash_id_db_filename)
            except InvalidHashIdDb:
                # The appropriate database is there but broken,
                # let's remove it and go on
                logging.warning("Invalid hash=>id database %s" %
                                hash_id_db_filename)
                os.remove(hash_id_db_filename)
                return

        result = self._determine_hash_id_db_filename()
        result.addCallback(use_it)
        return result

    def _determine_hash_id_db_filename(self):
        """Build up the filename of the hash=>id database to use.

        @return: a deferred resulting in the filename to use or C{None}
            in case of errors.
        """

        def got_server_uuid(server_uuid):

            warning = "Couldn't determine which hash=>id database to use: %s"

            if server_uuid is None:
                logging.warning(warning % "server UUID not available")
                return None

            try:
                lsb_release_info = parse_lsb_release(self.lsb_release_filename)
            except IOError, error:
                logging.warning(warning % str(error))
                return None
            try:
                codename = lsb_release_info["code-name"]
            except KeyError:
                logging.warning(warning % "missing code-name key in %s" %
                                self.lsb_release_filename)
                return None

            arch = self._facade.get_arch()
            if not arch:
                # The Apt code should always return a non-empty string,
                # so this branch shouldn't get executed at all. However
                # this check is kept as an extra paranoia sanity check.
                logging.warning(warning % "unknown dpkg architecture")
                return None

            return os.path.join(self._config.hash_id_directory,
                                "%s_%s_%s" % (server_uuid, codename, arch))

        result = self._broker.get_server_uuid()
        result.addCallback(got_server_uuid)
        return result

    def get_session_id(self):

        def got_session_id(session_id):
            self._session_id = session_id
            return session_id

        result = self._broker.get_session_id()
        result.addCallback(got_session_id)
        return result


def run_task_handler(cls, args, reactor=None):
    # please only pass reactor when you have totally mangled everything with
    # mocker. Otherwise bad things will happen.
    if reactor is None:
        reactor = LandscapeReactor()

    config = cls.config_factory()
    config.load(args)

    for directory in [config.package_directory, config.hash_id_directory]:
        if not os.path.isdir(directory):
            os.mkdir(directory)

    program_name = cls.queue_name
    lock_filename = os.path.join(config.package_directory,
                                 program_name + ".lock")
    try:
        lock_path(lock_filename)
    except LockError:
        if config.quiet:
            raise SystemExit()
        raise SystemExit("error: package %s is already running"
                         % program_name)

    words = re.findall("[A-Z][a-z]+", cls.__name__)
    init_logging(config, "-".join(word.lower() for word in words))

    # Setup our umask for Apt to use, this needs to setup file permissions to
    # 0644 so...
    os.umask(022)

    package_store = cls.package_store_class(config.store_filename)
    # Delay importing of the facades so that we don't
    # import Apt unless we need to.
    from landscape.package.facade import AptFacade
    package_facade = AptFacade()

    def finish():
        connector.disconnect()
        reactor.call_later(0, reactor.stop)

    def got_error(failure):
        log_failure(failure)
        finish()

    connector = RemoteBrokerConnector(reactor, config, retry_on_reconnect=True)
    remote = LazyRemoteBroker(connector)
    handler = cls(package_store, package_facade, remote, config)
    result = Deferred()
    result.addCallback(lambda x: handler.run())
    result.addCallback(lambda x: finish())
    result.addErrback(got_error)
    reactor.call_when_running(lambda: result.callback(None))
    reactor.run()

    return result



from __future__ import division

import os
import statvfs
import re


# List of filesystem types authorized when generating disk use statistics.
STABLE_FILESYSTEMS = frozenset(
    ["ext", "ext2", "ext3", "ext4", "reiserfs", "ntfs", "msdos", "dos", "vfat",
     "xfs", "hpfs", "jfs", "ufs", "hfs", "hfsplus"])


EXTRACT_DEVICE = re.compile("([a-z]+)[0-9]*")


def get_mount_info(mounts_file, statvfs_,
                   filesystems_whitelist=STABLE_FILESYSTEMS):
    """
    This is a generator that yields information about mounted filesystems.

    @param mounts_file: A file with information about mounted filesystems,
        such as C{/proc/mounts}.
    @param statvfs_: A function to get file status information.
    @param filesystems_whitelist: Optionally, a list of which filesystems to
        stat.
    @return: A C{dict} with C{device}, C{mount-point}, C{filesystem},
        C{total-space} and C{free-space} keys. If the filesystem information
        is not available, C{None} is returned. Both C{total-space} and
        C{free-space} are in megabytes.
    """
    for line in open(mounts_file):
        try:
            device, mount_point, filesystem = line.split()[:3]
            mount_point = mount_point.decode("string-escape")
        except ValueError:
            continue
        if (filesystems_whitelist is not None and
            filesystem not in filesystems_whitelist):
            continue
        megabytes = 1024 * 1024
        try:
            stats = statvfs_(mount_point)
        except OSError:
            continue
        block_size = stats[statvfs.F_BSIZE]
        total_space = (stats[statvfs.F_BLOCKS] * block_size) // megabytes
        free_space = (stats[statvfs.F_BFREE] * block_size) // megabytes
        yield {"device": device, "mount-point": mount_point,
               "filesystem": filesystem, "total-space": total_space,
               "free-space": free_space}


def get_filesystem_for_path(path, mounts_file, statvfs_):
    """
    Tries to determine to which of the mounted filesystem C{path} belongs to,
    and then returns information about that filesystem or C{None} if it
    couldn't be determined.

    @param path: The path we want filesystem information about.
    @param mounts_file: A file with information about mounted filesystems,
        such as C{/proc/mounts}.
    @param statvfs_: A function to get file status information.
    @param filesystems_whitelist: Optionally, a list of which filesystems to
        stat.
    @return: A C{dict} with C{device}, C{mount-point}, C{filesystem},
        C{total-space} and C{free-space} keys. If the filesystem information
        is not available, C{None} is returned. Both C{total-space} and
        C{free-space} are in megabytes.
    """
    candidate = None
    path = os.path.realpath(path)
    path_segments = path.split("/")
    for info in get_mount_info(mounts_file, statvfs_):
        mount_segments = info["mount-point"].split("/")
        if path.startswith(info["mount-point"]):
            if ((not candidate)
                or path_segments[:len(mount_segments)] == mount_segments):
                candidate = info
    return candidate


def is_device_removable(device):
    """
    This function returns whether a given device is removable or not by looking
    at the corresponding /sys/block/<device>/removable file

    @param device: The filesystem path to the device, e.g. /dev/sda1
    """
    # Shortcut the case where the device an SD card. The kernel/udev currently
    # consider SD cards (mmcblk devices) to be non-removable.
    if os.path.basename(device).startswith("mmcblk"):
        return True

    path = _get_device_removable_file_path(device)

    if not path:
        return False

    contents = None
    try:
        with open(path, "r") as f:
            contents = f.readline()
    except IOError:
        return False

    if contents.strip() == "1":
        return True
    return False


def _get_device_removable_file_path(device):
    """
    Get a device's "removable" file path.

    This function figures out the C{/sys/block/<device>/removable} path
    associated with the given device. The file at that path contains either
    a "0" if the device is not removable, or a "1" if it is.

    @param device: File system path of the device.
    """
    # The device will be a symlink if the disk is mounted by uuid or by label.
    if os.path.islink(device):
        # Paths are in the form "/dev/disk/by-uuid/<uuid>" and symlink
        # to the device file under /dev
        device = os.readlink(device)  # /dev/disk/by-uuid/<uuid> -> ../../sda1

    [device_name] = device.split("/")[-1:]  # /dev/sda1 -> sda1

    matched = EXTRACT_DEVICE.match(device_name)  # sda1 -> sda

    if not matched:
        return None

    device_name = matched.groups()[0]

    removable_file = os.path.join("/sys/block/", device_name, "removable")
    return removable_file



"""Tests for L{landscape.lib.fd}"""

import resource
from landscape.tests.mocker import ANY

from landscape.lib.fd import clean_fds
from landscape.tests.helpers import LandscapeTest


class CleanFDsTests(LandscapeTest):
    """Tests for L{clean_fds}."""

    def mock_rlimit(self, limit):
        getrlimit_mock = self.mocker.replace("resource.getrlimit")
        getrlimit_mock(resource.RLIMIT_NOFILE)
        self.mocker.result([None, limit])

    def test_clean_fds_rlimit(self):
        """
        L{clean_fds} cleans all non-stdio file descriptors up to the process
        limit for file descriptors.
        """
        self.mocker.order()
        self.mock_rlimit(10)
        close_mock = self.mocker.replace("os.close", passthrough=False)
        for i in range(3, 10):
            close_mock(i)
        self.mocker.replay()
        clean_fds()

    def test_clean_fds_sanity(self):
        """
        If the process limit for file descriptors is very high (> 4096), then
        we only close 4096 file descriptors.
        """
        self.mocker.order()
        self.mock_rlimit(4100)
        close_mock = self.mocker.replace("os.close", passthrough=False)
        closed_fds = []
        close_mock(ANY)
        self.mocker.call(closed_fds.append)
        self.mocker.count(4093)
        self.mocker.replay()
        clean_fds()
        self.assertEqual(closed_fds, range(3, 4096))

    def test_ignore_OSErrors(self):
        """
        If os.close raises an OSError, it is ignored and we continue to close
        the rest of the FDs.
        """
        self.mocker.order()
        self.mock_rlimit(10)

        closed_fds = []

        def remember_and_throw(fd):
            closed_fds.append(fd)
            raise OSError("Bad FD!")

        close_mock = self.mocker.replace("os.close", passthrough=False)
        close_mock(ANY)
        self.mocker.count(7)
        self.mocker.call(remember_and_throw)

        self.mocker.replay()
        clean_fds()
        self.assertEqual(closed_fds, range(3, 10))

    def test_dont_ignore_other_errors(self):
        """
        If other errors are raised from os.close, L{clean_fds} propagates them.
        """
        self.mocker.order()
        self.mock_rlimit(10)
        close_mock = self.mocker.replace("os.close", passthrough=False)
        close_mock(ANY)
        self.mocker.throw(MemoryError())

        self.mocker.replay()
        self.assertRaises(MemoryError, clean_fds)



import os


def detect_jiffies():
    """Returns the number of jiffies per second for this machine.

    A jiffy is a value used by the kernel to report certain time-based
    events.  Jiffies occur N times per second where N varies depending
    on the hardware the kernel is running on.  This function gets the
    uptime for the current process, forks a child process and gets the
    uptime again; finally, using the running time of the child process
    compared with the uptimes to determine number of jiffies per
    second.
    """
    uptime1_file = open("/proc/uptime")
    uptime2_file = open("/proc/uptime")
    read_uptime1 = uptime1_file.read
    read_uptime2 = uptime2_file.read

    while True:
        uptime1_data = read_uptime1()

        # Fork a process and exit immediately; this results in the
        # child process being left around as a zombie until waitpid()
        # is called.
        pid = os.fork()
        if pid == 0:
            os._exit(0)

        uptime2_data = read_uptime2()

        stat_file = open("/proc/%d/stat" % pid)
        stat_data = stat_file.read()
        stat_file.close()

        os.waitpid(pid, 0)

        seconds_uptime1 = float(uptime1_data.split()[0])
        seconds_uptime2 = float(uptime2_data.split()[0])
        jiffie_uptime = int(stat_data.split()[21])

        jiffies1 = int(jiffie_uptime/seconds_uptime1+0.5)
        jiffies2 = int(jiffie_uptime/seconds_uptime2+0.5)

        if jiffies1 == jiffies2:
            break

        uptime1_file.seek(0)
        uptime2_file.seek(0)

    uptime1_file.close()
    uptime2_file.close()
    return jiffies1



import unittest

from landscape.lib import bpickle


class BPickleTest(unittest.TestCase):

    def test_int(self):
        self.assertEqual(bpickle.loads(bpickle.dumps(1)), 1)

    def test_float(self):
        self.assertAlmostEquals(bpickle.loads(bpickle.dumps(2.3)), 2.3)

    def test_float_scientific_notation(self):
        number = 0.00005
        self.assertTrue("e" in repr(number))
        self.assertAlmostEquals(bpickle.loads(bpickle.dumps(number)), number)

    def test_string(self):
        self.assertEqual(bpickle.loads(bpickle.dumps('foo')), 'foo')

    def test_list(self):
        self.assertEqual(bpickle.loads(bpickle.dumps([1, 2, 'hello', 3.0])),
                         [1, 2, 'hello', 3.0])

    def test_tuple(self):
        data = bpickle.dumps((1, [], 2, 'hello', 3.0))
        self.assertEqual(bpickle.loads(data),
                         (1, [], 2, 'hello', 3.0))

    def test_none(self):
        self.assertEqual(bpickle.loads(bpickle.dumps(None)), None)

    def test_unicode(self):
        self.assertEqual(bpickle.loads(bpickle.dumps(u'\xc0')), u'\xc0')

    def test_bool(self):
        self.assertEqual(bpickle.loads(bpickle.dumps(True)), True)

    def test_dict(self):
        dumped_tostr = bpickle.dumps({True: "hello"})
        self.assertEqual(bpickle.loads(dumped_tostr),
                         {True: "hello"})
        dumped_tobool = bpickle.dumps({True: False})
        self.assertEqual(bpickle.loads(dumped_tobool),
                         {True: False})

    def test_long(self):
        long = 99999999999999999999999999999
        self.assertEqual(bpickle.loads(bpickle.dumps(long)), long)



import os
import re

from twisted.internet.defer import succeed, fail, inlineCallbacks

from landscape.lib.fetch import HTTPCodeError, PyCurlError
from landscape.lib.fs import create_file
from landscape.monitor.computerinfo import ComputerInfo, METADATA_RETRY_MAX
from landscape.tests.helpers import LandscapeTest, MonitorHelper
from landscape.tests.mocker import ANY

SAMPLE_LSB_RELEASE = "DISTRIB_ID=Ubuntu\n"                         \
                     "DISTRIB_RELEASE=6.06\n"                      \
                     "DISTRIB_CODENAME=dapper\n"                   \
                     "DISTRIB_DESCRIPTION=\"Ubuntu 6.06.1 LTS\"\n"


def get_fqdn():
    return "ooga.local"


class ComputerInfoTest(LandscapeTest):

    helpers = [MonitorHelper]

    sample_memory_info = """
MemTotal:      1547072 kB
MemFree:        106616 kB
Buffers:        267088 kB
Cached:         798388 kB
SwapCached:          0 kB
Active:         728952 kB
Inactive:       536512 kB
HighTotal:      646016 kB
HighFree:        42204 kB
LowTotal:       901056 kB
LowFree:         64412 kB
SwapTotal:     1622524 kB
SwapFree:      1622524 kB
Dirty:              24 kB
Writeback:           0 kB
Mapped:         268756 kB
Slab:           105492 kB
CommitLimit:   2396060 kB
Committed_AS:  1166936 kB
PageTables:       2748 kB
VmallocTotal:   114680 kB
VmallocUsed:      6912 kB
VmallocChunk:   107432 kB
"""

    def setUp(self):
        LandscapeTest.setUp(self)
        self.lsb_release_filename = self.makeFile(SAMPLE_LSB_RELEASE)
        self.query_results = {}

        def fetch_stub(url, **kwargs):
            value = self.query_results[url]
            if isinstance(value, Exception):
                return fail(value)
            else:
                return succeed(value)

        self.fetch_func = fetch_stub
        self.add_query_result("instance-id", "i00001")
        self.add_query_result("ami-id", "ami-00002")
        self.add_query_result("instance-type", "hs1.8xlarge")

    def add_query_result(self, name, value):
        """
        Add a url to self.query_results that is then available through
        self.fetch_func.
        """
        url = "http://169.254.169.254/latest/meta-data/" + name
        self.query_results[url] = value

    def test_get_fqdn(self):
        self.mstore.set_accepted_types(["computer-info"])
        plugin = ComputerInfo(get_fqdn=get_fqdn, fetch_async=self.fetch_func)
        self.monitor.add(plugin)
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0]["type"], "computer-info")
        self.assertEqual(messages[0]["hostname"], "ooga.local")

    def test_get_real_hostname(self):
        self.mstore.set_accepted_types(["computer-info"])
        plugin = ComputerInfo(fetch_async=self.fetch_func)
        self.monitor.add(plugin)
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0]["type"], "computer-info")
        self.assertNotEquals(len(messages[0]["hostname"]), 0)
        self.assertTrue(re.search("\w", messages[0]["hostname"]))

    def test_only_report_changed_hostnames(self):
        self.mstore.set_accepted_types(["computer-info"])
        plugin = ComputerInfo(get_fqdn=get_fqdn, fetch_async=self.fetch_func)
        self.monitor.add(plugin)
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)

    def test_report_changed_hostnames(self):

        def hostname_factory(hostnames=["ooga", "wubble", "wubble"]):
            i = 0
            while i < len(hostnames):
                yield hostnames[i]
                i = i + 1

        self.mstore.set_accepted_types(["computer-info"])
        plugin = ComputerInfo(get_fqdn=hostname_factory().next,
                              fetch_async=self.fetch_func)
        self.monitor.add(plugin)

        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0]["hostname"], "ooga")

        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 2)
        self.assertEqual(messages[1]["hostname"], "wubble")

    def test_get_total_memory(self):
        self.mstore.set_accepted_types(["computer-info"])
        meminfo_filename = self.makeFile(self.sample_memory_info)
        plugin = ComputerInfo(meminfo_filename=meminfo_filename,
                              fetch_async=self.fetch_func)
        self.monitor.add(plugin)
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        message = messages[0]
        self.assertEqual(message["type"], "computer-info")
        self.assertEqual(message["total-memory"], 1510)
        self.assertEqual(message["total-swap"], 1584)

    def test_get_real_total_memory(self):
        self.mstore.set_accepted_types(["computer-info"])
        self.makeFile(self.sample_memory_info)
        plugin = ComputerInfo(fetch_async=self.fetch_func)
        self.monitor.add(plugin)
        plugin.exchange()
        message = self.mstore.get_pending_messages()[0]
        self.assertEqual(message["type"], "computer-info")
        self.assertTrue(isinstance(message["total-memory"], int))
        self.assertTrue(isinstance(message["total-swap"], int))

    def test_wb_report_changed_total_memory(self):
        self.mstore.set_accepted_types(["computer-info"])
        plugin = ComputerInfo(fetch_async=self.fetch_func)
        self.monitor.add(plugin)

        plugin._get_memory_info = lambda: (1510, 1584)
        plugin.exchange()
        message = self.mstore.get_pending_messages()[0]
        self.assertEqual(message["total-memory"], 1510)
        self.assertTrue("total-swap" in message)

        plugin._get_memory_info = lambda: (2048, 1584)
        plugin.exchange()
        message = self.mstore.get_pending_messages()[1]
        self.assertEqual(message["total-memory"], 2048)
        self.assertTrue("total-swap" not in message)

    def test_wb_report_changed_total_swap(self):
        self.mstore.set_accepted_types(["computer-info"])
        plugin = ComputerInfo(fetch_async=self.fetch_func)
        self.monitor.add(plugin)

        plugin._get_memory_info = lambda: (1510, 1584)
        plugin.exchange()
        message = self.mstore.get_pending_messages()[0]
        self.assertEqual(message["total-swap"], 1584)
        self.assertTrue("total-memory" in message)

        plugin._get_memory_info = lambda: (1510, 2048)
        plugin.exchange()
        message = self.mstore.get_pending_messages()[1]
        self.assertEqual(message["total-swap"], 2048)
        self.assertTrue("total-memory" not in message)

    def test_get_distribution(self):
        """
        Various details about the distribution should be reported by
        the plugin.  This test ensures that the right kinds of details
        end up in messages produced by the plugin.
        """
        self.mstore.set_accepted_types(["distribution-info"])
        plugin = ComputerInfo()
        self.monitor.add(plugin)

        plugin.exchange()
        message = self.mstore.get_pending_messages()[0]
        self.assertEqual(message["type"], "distribution-info")
        self.assertTrue("distributor-id" in message)
        self.assertTrue("description" in message)
        self.assertTrue("release" in message)
        self.assertTrue("code-name" in message)

    def test_get_sample_distribution(self):
        """
        Sample data is used to ensure that expected values end up in
        the distribution data reported by the plugin.
        """
        self.mstore.set_accepted_types(["distribution-info"])
        plugin = ComputerInfo(lsb_release_filename=self.lsb_release_filename)
        self.monitor.add(plugin)

        plugin.exchange()
        message = self.mstore.get_pending_messages()[0]
        self.assertEqual(message["type"], "distribution-info")
        self.assertEqual(message["distributor-id"], "Ubuntu")
        self.assertEqual(message["description"], "Ubuntu 6.06.1 LTS")
        self.assertEqual(message["release"], "6.06")
        self.assertEqual(message["code-name"], "dapper")

    def test_distribution_reported_only_once(self):
        """
        Distribution data shouldn't be reported unless it's changed
        since the last time it was reported.
        """
        self.mstore.set_accepted_types(["distribution-info"])
        plugin = ComputerInfo()
        self.monitor.add(plugin)

        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0]["type"], "distribution-info")

        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)

    def test_wb_report_changed_distribution(self):
        """
        When distribution data changes, the new data should be sent to
        the server.
        """
        self.mstore.set_accepted_types(["distribution-info"])
        plugin = ComputerInfo(lsb_release_filename=self.lsb_release_filename)
        self.monitor.add(plugin)

        plugin.exchange()
        message = self.mstore.get_pending_messages()[0]
        self.assertEqual(message["type"], "distribution-info")
        self.assertEqual(message["distributor-id"], "Ubuntu")
        self.assertEqual(message["description"], "Ubuntu 6.06.1 LTS")
        self.assertEqual(message["release"], "6.06")
        self.assertEqual(message["code-name"], "dapper")

        plugin._lsb_release_filename = self.makeFile("""\
DISTRIB_ID=Ubuntu
DISTRIB_RELEASE=6.10
DISTRIB_CODENAME=edgy
DISTRIB_DESCRIPTION="Ubuntu 6.10"
""")
        plugin.exchange()
        message = self.mstore.get_pending_messages()[1]
        self.assertEqual(message["type"], "distribution-info")
        self.assertEqual(message["distributor-id"], "Ubuntu")
        self.assertEqual(message["description"], "Ubuntu 6.10")
        self.assertEqual(message["release"], "6.10")
        self.assertEqual(message["code-name"], "edgy")

    def test_unknown_distribution_key(self):
        self.mstore.set_accepted_types(["distribution-info"])
        lsb_release_filename = self.makeFile("""\
DISTRIB_ID=Ubuntu
DISTRIB_RELEASE=6.10
DISTRIB_CODENAME=edgy
DISTRIB_DESCRIPTION="Ubuntu 6.10"
DISTRIB_NEW_UNEXPECTED_KEY=ooga
""")
        plugin = ComputerInfo(lsb_release_filename=lsb_release_filename)
        self.monitor.add(plugin)

        plugin.exchange()
        message = self.mstore.get_pending_messages()[0]
        self.assertEqual(message["type"], "distribution-info")
        self.assertEqual(message["distributor-id"], "Ubuntu")
        self.assertEqual(message["description"], "Ubuntu 6.10")
        self.assertEqual(message["release"], "6.10")
        self.assertEqual(message["code-name"], "edgy")

    def test_resynchronize(self):
        """
        If a reactor event "resynchronize" is received, messages for
        all computer info should be generated.
        """
        self.mstore.set_accepted_types(["distribution-info", "computer-info"])
        meminfo_filename = self.makeFile(self.sample_memory_info)
        plugin = ComputerInfo(get_fqdn=get_fqdn,
                              meminfo_filename=meminfo_filename,
                              lsb_release_filename=self.lsb_release_filename,
                              root_path=self.makeDir(),
                              fetch_async=self.fetch_func)
        self.monitor.add(plugin)
        plugin.exchange()
        self.reactor.fire("resynchronize", scopes=["computer"])
        plugin.exchange()
        computer_info = {"type": "computer-info", "hostname": "ooga.local",
                         "timestamp": 0, "total-memory": 1510,
                         "total-swap": 1584}

        dist_info = {"type": "distribution-info",
                     "code-name": "dapper", "description": "Ubuntu 6.06.1 LTS",
                     "distributor-id": "Ubuntu", "release": "6.06"}
        self.assertMessages(self.mstore.get_pending_messages(),
                            [computer_info, dist_info,
                             computer_info, dist_info])

    def test_computer_info_call_on_accepted(self):
        plugin = ComputerInfo(fetch_async=self.fetch_func)
        self.monitor.add(plugin)

        remote_broker_mock = self.mocker.replace(self.remote)
        remote_broker_mock.send_message(ANY, ANY, urgent=True)
        self.mocker.replay()

        self.mstore.set_accepted_types(["computer-info"])
        self.reactor.fire(("message-type-acceptance-changed", "computer-info"),
                          True)

    def test_distribution_info_call_on_accepted(self):
        plugin = ComputerInfo()
        self.monitor.add(plugin)

        remote_broker_mock = self.mocker.replace(self.remote)
        remote_broker_mock.send_message(ANY, ANY, urgent=True)
        self.mocker.replay()

        self.mstore.set_accepted_types(["distribution-info"])
        self.reactor.fire(("message-type-acceptance-changed",
                           "distribution-info"),
                          True)

    def test_message_if_not_accepted(self):
        """
        Don't add any messages at all if the broker isn't currently
        accepting their type.
        """
        plugin = ComputerInfo()
        self.monitor.add(plugin)
        self.reactor.advance(self.monitor.step_size * 2)
        self.monitor.exchange()

        self.mstore.set_accepted_types(["distribution-info", "computer-info"])
        self.assertMessages(list(self.mstore.get_pending_messages()), [])

    def test_annotations(self):
        """
        L{ComputerInfo} sends extra meta data from the annotations.d directory
        if it's present.

        Each file name is used as a key in the meta-data dict and the file's
        contents are used as values.

        This allows system administrators to add extra per-computer information
        into Landscape which makes sense to them.
        """
        annotations_dir = self.monitor.config.annotations_path
        os.mkdir(annotations_dir)
        create_file(os.path.join(annotations_dir, "annotation1"), "value1")
        create_file(os.path.join(annotations_dir, "annotation2"), "value2")
        self.mstore.set_accepted_types(["computer-info"])

        plugin = ComputerInfo()
        plugin._cloud_meta_data = {}
        self.monitor.add(plugin)
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(1, len(messages))
        meta_data = messages[0]["annotations"]
        self.assertEqual(2, len(meta_data))
        self.assertEqual("value1", meta_data["annotation1"])
        self.assertEqual("value2", meta_data["annotation2"])

    def test_fetch_cloud_metadata(self):
        """
        Fetch cloud information and insert it in a cloud-instance-metadata
        message.
        """
        self.config.cloud = True
        self.mstore.set_accepted_types(["cloud-instance-metadata"])

        plugin = ComputerInfo(fetch_async=self.fetch_func)
        self.monitor.add(plugin)
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(1, len(messages))
        self.assertEqual(u"i00001", messages[0]["instance-id"])
        self.assertEqual(u"ami-00002", messages[0]["ami-id"])
        self.assertEqual(u"hs1.8xlarge", messages[0]["instance-type"])

    def test_send_cloud_instance_metadata_only_once(self):
        """Only send the cloud information once per client restart."""
        self.config.cloud = True
        self.mstore.set_accepted_types(["cloud-instance-metadata"])

        plugin = ComputerInfo(fetch_async=self.fetch_func)
        self.monitor.add(plugin)
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(1, len(messages))
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(1, len(messages))

    def test_no_fetch_ec2_meta_data_when_cloud_retries_is_max(self):
        """
        Do not fetch EC2 info when C{_cloud_retries} is C{METADATA_RETRY_MAX}
        """
        self.config.cloud = True
        self.mstore.set_accepted_types(["cloud-instance-metadata"])

        plugin = ComputerInfo(fetch_async=self.fetch_func)
        plugin._cloud_retries = METADATA_RETRY_MAX
        self.monitor.add(plugin)
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(0, len(messages))

    @inlineCallbacks
    def test_fetch_ec2_meta_data(self):
        """
        L{_fetch_ec2_meta_data} retrieves instance information from the
        EC2 api.
        """
        plugin = ComputerInfo(fetch_async=self.fetch_func)
        result = yield plugin._fetch_ec2_meta_data()
        self.assertEqual({"instance-id": u"i00001", "ami-id": u"ami-00002",
                          "instance-type": u"hs1.8xlarge"}, result)
        self.assertEqual(
            "    INFO: Querying cloud meta-data.\n"
            "    INFO: Acquired cloud meta-data.\n",
            self.logfile.getvalue())

    @inlineCallbacks
    def test_fetch_ec2_meta_data_no_cloud_api_max_retry(self):
        """
        L{_fetch_ec2_meta_data} returns C{None} when faced with no EC2 cloud
        API service and reports the specific C{PyCurlError} upon message
        exchange when L{_cloud_retries} equals C{METADATA_RETRY_MAX}.
        """
        self.log_helper.ignore_errors(PyCurlError)
        self.add_query_result("instance-id", PyCurlError(60, "pycurl error"))
        plugin = ComputerInfo(fetch_async=self.fetch_func)
        plugin._cloud_retries = METADATA_RETRY_MAX
        result = yield plugin._fetch_ec2_meta_data()
        self.assertIn(
            "INFO: No cloud meta-data available. "
            "Error 60: pycurl error\n", self.logfile.getvalue())
        self.assertEqual(None, result)

    @inlineCallbacks
    def test_fetch_ec2_meta_data_bad_result_max_retry(self):
        """
        L{_fetch_ec2_meta_data} returns C{None} and logs an error when
        crossing the retry threshold C{METADATA_RETRY_MAX}.
        """
        self.log_helper.ignore_errors(HTTPCodeError)
        self.add_query_result("ami-id", HTTPCodeError(404, "notfound"))
        plugin = ComputerInfo(fetch_async=self.fetch_func)
        plugin._cloud_retries = METADATA_RETRY_MAX
        result = yield plugin._fetch_ec2_meta_data()
        self.assertIn(
            "INFO: No cloud meta-data available. Server returned "
            "HTTP code 404",
            self.logfile.getvalue())
        self.assertEqual(None, result)

    @inlineCallbacks
    def test_fetch_ec2_meta_data_bad_result_retry(self):
        """
        L{_fetch_ec2_meta_data} returns C{None} when faced with spurious
        errors from the EC2 api. The method increments L{_cloud_retries}
        counter which allows L{_fetch_ec2_meta_data} to run again next
        message exchange.
        """
        self.log_helper.ignore_errors(HTTPCodeError)
        self.add_query_result("ami-id", HTTPCodeError(404, "notfound"))
        plugin = ComputerInfo(fetch_async=self.fetch_func)
        result = yield plugin._fetch_ec2_meta_data()
        self.assertEqual(1, plugin._cloud_retries)
        self.assertEqual(None, result)
        # Fix the error condition for the retry.
        self.add_query_result("ami-id", "ami-00002")
        result = yield plugin._fetch_ec2_meta_data()
        self.assertEqual({"instance-id": u"i00001", "ami-id": u"ami-00002",
                          "instance-type": u"hs1.8xlarge"},
                         result)



import os

from twisted.internet import reactor
from twisted.internet.defer import Deferred

from landscape.tests.helpers import LandscapeTest
from landscape.lib.gpg import gpg_verify


class GPGTest(LandscapeTest):

    def test_gpg_verify(self):
        """
        L{gpg_verify} runs the given gpg binary and returns C{True} if the
        provided signature is valid.
        """
        gpg_options = self.makeFile()
        gpg = self.makeFile("#!/bin/sh\n"
                            "touch $3/trustdb.gpg\n"
                            "echo -n $@ > %s\n" % gpg_options)
        os.chmod(gpg, 0755)
        gpg_home = self.makeDir()
        mkdtemp_mock = self.mocker.replace("tempfile.mkdtemp")
        mkdtemp_mock()
        self.mocker.result(gpg_home)
        self.mocker.replay()

        deferred = Deferred()

        def do_test():

            result = gpg_verify("/some/file", "/some/signature", gpg=gpg)

            def check_result(ignored):
                self.assertEqual(
                    open(gpg_options).read(),
                    "--no-options --homedir %s --no-default-keyring "
                    "--ignore-time-conflict --keyring /etc/apt/trusted.gpg "
                    "--verify /some/signature /some/file" % gpg_home)
                self.assertFalse(os.path.exists(gpg_home))

            result.addCallback(check_result)
            result.chainDeferred(deferred)

        reactor.callWhenRunning(do_test)
        return deferred

    def test_gpg_verify_with_non_zero_gpg_exit_code(self):
        """
        L{gpg_verify} runs the given gpg binary and returns C{False} if the
        provided signature is not valid.
        """
        gpg = self.makeFile("#!/bin/sh\necho out; echo err >&2; exit 1\n")
        os.chmod(gpg, 0755)
        gpg_home = self.makeDir()
        mkdtemp_mock = self.mocker.replace("tempfile.mkdtemp")
        mkdtemp_mock()
        self.mocker.result(gpg_home)
        self.mocker.replay()

        deferred = Deferred()

        def do_test():

            result = gpg_verify("/some/file", "/some/signature", gpg=gpg)

            def check_failure(failure):
                self.assertEqual(str(failure.value),
                                 "%s failed (out='out\n', err='err\n', "
                                 "code='1')" % gpg)
                self.assertFalse(os.path.exists(gpg_home))

            result.addCallback(self.fail)
            result.addErrback(check_failure)
            result.chainDeferred(deferred)

        reactor.callWhenRunning(do_test)
        return deferred



import os

from landscape.tests.helpers import LandscapeTest
from landscape.lib.twisted_util import spawn_process
from landscape.lib.fs import create_file


class SpawnProcessTest(LandscapeTest):

    def setUp(self):
        super(SpawnProcessTest, self).setUp()
        self.command = self.makeFile("#!/bin/sh\necho -n $@")
        os.chmod(self.command, 0755)

    def test_spawn_process_return_value(self):
        """
        The process is executed and returns the expected exit code.
        """
        create_file(self.command, "#!/bin/sh\nexit 2")

        def callback((out, err, code)):
            self.assertEqual(out, "")
            self.assertEqual(err, "")
            self.assertEqual(code, 2)

        result = spawn_process(self.command)
        result.addCallback(callback)
        return result

    def test_spawn_process_output(self):
        """
        The process returns the expected standard output.
        """
        def callback((out, err, code)):
            self.assertEqual(out, "a b")
            self.assertEqual(err, "")
            self.assertEqual(code, 0)

        result = spawn_process(self.command, args=("a", "b"))
        result.addCallback(callback)
        return result

    def test_spawn_process_error(self):
        """
        The process returns the expected standard error.
        """
        create_file(self.command, "#!/bin/sh\necho -n $@ >&2")

        def callback((out, err, code)):
            self.assertEqual(out, "")
            self.assertEqual(err, "a b")
            self.assertEqual(code, 0)

        result = spawn_process(self.command, args=("a", "b"))
        result.addCallback(callback)
        return result

    def test_spawn_process_callback(self):
        """
        If a callback for process output is provieded, it is called for every
        line of output.
        """
        create_file(self.command, "#!/bin/sh\n/bin/echo -ne $@")
        param = r"some text\nanother line\nok, last one\n"
        expected = ["some text", "another line", "ok, last one"]
        lines = []

        def line_received(line):
            lines.append(line)

        def callback((out, err, code)):
            self.assertEqual(expected, lines)

        result = spawn_process(self.command, args=(param,),
                               line_received=line_received)
        result.addCallback(callback)
        return result

    def test_spawn_process_callback_multiple_newlines(self):
        """
        If output ends with more than one newline, empty lines are preserved.
        """
        create_file(self.command, "#!/bin/sh\n/bin/echo -ne $@")
        param = r"some text\nanother line\n\n\n"
        expected = ["some text", "another line", "", ""]
        lines = []

        def line_received(line):
            lines.append(line)

        def callback((out, err, code)):
            self.assertEqual(expected, lines)

        result = spawn_process(self.command, args=(param,),
                               line_received=line_received)
        result.addCallback(callback)
        return result

    def test_spawn_process_callback_no_newline(self):
        """
        If output ends without a newline, the line is still passed to the
        callback.
        """
        create_file(self.command, "#!/bin/sh\n/bin/echo -ne $@")
        param = r"some text\nanother line\nok, last one"
        expected = ["some text", "another line", "ok, last one"]
        lines = []

        def line_received(line):
            lines.append(line)

        def callback((out, err, code)):
            self.assertEqual(expected, lines)

        result = spawn_process(self.command, args=(param,),
                               line_received=line_received)
        result.addCallback(callback)
        return result

    def test_spawn_process_with_stdin(self):
        """
        Optionally C{spawn_process} accepts a C{stdin} argument.
        """
        create_file(self.command, "#!/bin/sh\n/bin/cat")

        def callback((out, err, code)):
            self.assertEqual("hello", out)

        result = spawn_process(self.command, stdin="hello")
        result.addCallback(callback)
        return result



from landscape.monitor.loadaverage import LoadAverage
from landscape.tests.helpers import LandscapeTest, MonitorHelper
from landscape.tests.mocker import ANY


def get_load_average():
    i = 1
    while True:
        yield (float(i), 1.0, 500.238)
        i += 1


class LoadAveragePluginTest(LandscapeTest):

    helpers = [MonitorHelper]

    def test_real_load_average(self):
        """
        When the load average plugin runs it calls os.getloadavg() to
        retrieve current load average data.  This test makes sure that
        os.getloadavg() is called without failing and that messages
        with the expected datatypes are generated.
        """
        plugin = LoadAverage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        self.reactor.advance(self.monitor.step_size)

        message = plugin.create_message()
        self.assertTrue("type" in message)
        self.assertEqual(message["type"], "load-average")
        self.assertTrue("load-averages" in message)

        load_averages = message["load-averages"]
        self.assertEqual(len(load_averages), 1)

        load_average = load_averages[0]
        self.assertEqual(load_average[0], self.monitor.step_size)
        self.assertTrue(isinstance(load_average[0], int))
        self.assertTrue(isinstance(load_average[1], float))

    def test_sample_load_average(self):
        """
        Sample data is used to ensure that the load average included
        in the message is calculated correctly.
        """
        get_load_average = lambda: (0.15, 1, 500)
        plugin = LoadAverage(create_time=self.reactor.time,
                             get_load_average=get_load_average)
        self.monitor.add(plugin)

        self.reactor.advance(self.monitor.step_size)

        message = plugin.create_message()
        load_averages = message["load-averages"]
        self.assertEqual(len(load_averages), 1)
        self.assertEqual(load_averages[0], (self.monitor.step_size, 0.15))

    def test_ranges_remain_contiguous_after_flush(self):
        """
        The load average plugin uses the accumulate function to queue
        messages.  Timestamps should always be contiguous, and always
        fall on a step boundary.
        """
        plugin = LoadAverage(create_time=self.reactor.time,
                             get_load_average=get_load_average().next)
        self.monitor.add(plugin)

        for i in range(1, 10):
            self.reactor.advance(self.monitor.step_size)
            message = plugin.create_message()
            load_averages = message["load-averages"]
            self.assertEqual(len(load_averages), 1)
            self.assertEqual(load_averages[0][0], self.monitor.step_size * i)

    def test_messaging_flushes(self):
        """
        Duplicate message should never be created.  If no data is
        available, a message with an empty C{load-averages} list is
        expected.
        """
        plugin = LoadAverage(create_time=self.reactor.time,
                             get_load_average=get_load_average().next)
        self.monitor.add(plugin)

        self.reactor.advance(self.monitor.step_size)

        message = plugin.create_message()
        self.assertEqual(len(message["load-averages"]), 1)

        message = plugin.create_message()
        self.assertEqual(len(message["load-averages"]), 0)

    def test_never_exchange_empty_messages(self):
        """
        The plugin will create a message with an empty
        C{load-averages} list when no data is available.  If an empty
        message is created during exchange, it should not be queued.
        """
        self.mstore.set_accepted_types(["load-average"])

        plugin = LoadAverage(create_time=self.reactor.time,
                             get_load_average=get_load_average().next)
        self.monitor.add(plugin)

        self.monitor.exchange()
        self.assertEqual(len(self.mstore.get_pending_messages()), 0)

    def test_exchange_messages(self):
        """
        The load average plugin queues message when manager.exchange()
        is called.  Each message should be aligned to a step boundary;
        messages collected bewteen exchange periods should be
        delivered in a single message.
        """
        self.mstore.set_accepted_types(["load-average"])

        plugin = LoadAverage(create_time=self.reactor.time,
                             get_load_average=get_load_average().next)
        self.monitor.add(plugin)

        self.reactor.advance(self.monitor.step_size * 2)
        self.monitor.exchange()

        self.assertMessages(self.mstore.get_pending_messages(),
                            [{"type": "load-average",
                              "load-averages": [(300, 10.5), (600, 30.5)]}])

    def test_call_on_accepted(self):
        plugin = LoadAverage(create_time=self.reactor.time,
                             get_load_average=get_load_average().next)
        self.monitor.add(plugin)

        self.reactor.advance(self.monitor.step_size * 1)

        remote_broker_mock = self.mocker.replace(self.remote)
        remote_broker_mock.send_message(ANY, ANY, urgent=True)
        self.mocker.replay()

        self.reactor.fire(("message-type-acceptance-changed", "load-average"),
                          True)

    def test_no_message_if_not_accepted(self):
        """
        Don't add any messages at all if the broker isn't currently
        accepting their type.
        """
        plugin = LoadAverage(create_time=self.reactor.time,
                             get_load_average=get_load_average().next)
        self.monitor.add(plugin)

        self.reactor.advance(self.monitor.step_size * 2)
        self.monitor.exchange()

        self.mstore.set_accepted_types(["load-average"])
        self.assertMessages(list(self.mstore.get_pending_messages()), [])



import logging
import os

from twisted.internet.utils import getProcessOutput
from twisted.internet.defer import succeed

from landscape.package.store import PackageStore
from landscape.package.changer import PackageChanger
from landscape.package.releaseupgrader import ReleaseUpgrader
from landscape.manager.plugin import ManagerPlugin


class PackageManager(ManagerPlugin):

    run_interval = 1800
    _package_store = None

    def register(self, registry):
        super(PackageManager, self).register(registry)
        self.config = registry.config

        if not self._package_store:
            filename = os.path.join(registry.config.data_path,
                                    "package/database")
            self._package_store = PackageStore(filename)

        registry.register_message("change-packages",
                                  self.handle_change_packages)
        registry.register_message("change-package-locks",
                                  self.handle_change_package_locks)
        registry.register_message("release-upgrade",
                                  self.handle_release_upgrade)

        # When the package reporter notifies us that something has changed,
        # we want to run again to see if we can now fulfill tasks that were
        # skipped before.
        registry.reactor.call_on("package-data-changed", self.run)

        self.run()

    def _handle(self, cls, message):
        """Queue C{message} as a task, and spawn the proper handler."""
        self._package_store.add_task(cls.queue_name, message)
        self.spawn_handler(cls)

    def handle_change_packages(self, message):
        return self._handle(PackageChanger, message)

    def handle_change_package_locks(self, message):
        return self._handle(PackageChanger, message)

    def handle_release_upgrade(self, message):
        return self._handle(ReleaseUpgrader, message)

    def run(self):
        result = self.registry.broker.get_accepted_message_types()
        result.addCallback(self._got_message_types)
        return result

    def _got_message_types(self, message_types):
        if "change-packages-result" in message_types:
            self.spawn_handler(PackageChanger)
        if "operation-result" in message_types:
            self.spawn_handler(ReleaseUpgrader)

    def spawn_handler(self, cls):
        args = ["--quiet"]
        if self.config.config:
            args.extend(["-c", self.config.config])
        if self._package_store.get_next_task(cls.queue_name):
            # path is set to None so that getProcessOutput does not
            # chdir to "." see bug #211373
            result = getProcessOutput(cls.find_command(),
                                      args=args, env=os.environ,
                                      errortoo=1,
                                      path=None)
            result.addCallback(self._got_output, cls)
        else:
            result = succeed(None)
        return result

    def _got_output(self, output, cls):
        if output:
            logging.warning("Package %s output:\n%s" %
                            (cls.queue_name, output))



import ConfigParser
import os
import logging

from landscape.monitor.plugin import MonitorPlugin


class UpdateManager(MonitorPlugin):
    """
    Report on changes to the update-manager configuration.

    @param update_manager_filename: the path to the update-manager
        configuration file.
    """

    # This file is used by the update-manager and may contain a "Prompt"
    # variable which indicates that users are prompted to upgrade the release
    # when any new release is available ("normal"); when a new LTS release is
    # available ("lts"); or never ("never").
    update_manager_filename = "/etc/update-manager/release-upgrades"

    persist_name = "update-manager"
    scope = "package"
    run_interval = 3600  # 1 hour
    run_immediately = True

    def __init__(self, update_manager_filename=None):
        if update_manager_filename is not None:
            self.update_manager_filename = update_manager_filename

    def _get_prompt(self):
        """
        Retrieve the update-manager upgrade prompt which dictates when we
        should prompt users to upgrade the release.  Current valid values are
        "normal" (prompt on all the availability of all releases), "lts"
        (prompt only when LTS releases are available), and "never".
        """
        if not os.path.exists(self.update_manager_filename):
            # There is no config, so we just act as if it's set to 'normal'
            return "normal"
        config_file = open(self.update_manager_filename)
        parser = ConfigParser.SafeConfigParser()
        parser.readfp(config_file)
        prompt = parser.get("DEFAULT", "Prompt")
        valid_prompts = ["lts", "never", "normal"]
        if prompt not in valid_prompts:
            prompt = "normal"
            message = ("%s contains invalid Prompt value. "
                       "Should be one of %s." % (
                           self.update_manager_filename,
                           valid_prompts))
            logging.warning(message)
        return prompt

    def send_message(self):
        """
        Send the current upgrade release prompt to the server.
        """
        prompt = self._get_prompt()
        if prompt == self._persist.get("prompt"):
            return
        self._persist.set("prompt", prompt)
        message = {
            "type": "update-manager-info",
            "prompt": prompt}
        logging.info("Queueing message with updated "
                     "update-manager status.")
        return self.registry.broker.send_message(message, self._session_id)

    def run(self):
        """
        Send the update-manager-info messages, if the server accepts them.
        """
        return self.registry.broker.call_if_accepted(
            "update-manager-info", self.send_message)



from landscape.tests.helpers import LandscapeTest

from twisted.internet.defer import fail

from landscape.lib.bpickle import dumps
from landscape.lib.fetch import fetch
from landscape.reactor import FakeReactor
from landscape.broker.ping import PingClient, Pinger
from landscape.broker.tests.helpers import ExchangeHelper


class FakePageGetter(object):
    """An fake web client."""

    def __init__(self, response):
        self.response = response
        self.fetches = []

    def get_page(self, url, post, headers, data):
        """
        A method which is supposed to act like a limited version of
        L{landscape.lib.fetch.fetch}.

        Record attempts to get pages, and return a deferred with pre-cooked
        data.
        """
        self.fetches.append((url, post, headers, data))
        return dumps(self.response)

    def failing_get_page(self, url, post, headers, data):
        """
        A method which is supposed to act like a limited version of
        L{landscape.lib.fetch.fetch}.

        Record attempts to get pages, and return a deferred with pre-cooked
        data.
        """
        raise AssertionError("That's a failure!")


class PingClientTest(LandscapeTest):

    def setUp(self):
        super(PingClientTest, self).setUp()
        self.reactor = FakeReactor()

    def test_default_get_page(self):
        """
        The C{get_page} argument to L{PingClient} should be optional, and
        default to L{twisted.web.client.getPage}.
        """
        client = PingClient(self.reactor)
        self.assertEqual(client.get_page, fetch)

    def test_ping(self):
        """
        L{PingClient} should be able to send a web request to a specified URL
        about a particular insecure ID.
        """
        client = FakePageGetter(None)
        url = "http://localhost/ping"
        insecure_id = 10
        pinger = PingClient(self.reactor, get_page=client.get_page)
        pinger.ping(url, insecure_id)
        self.assertEqual(
            client.fetches,
            [(url, True, {"Content-Type": "application/x-www-form-urlencoded"},
              "insecure_id=10")])

    def test_ping_no_insecure_id(self):
        """
        If a L{PingClient} does not have an insecure-id yet, then the ping
        should not happen.
        """
        client = FakePageGetter(None)
        pinger = PingClient(self.reactor, get_page=client.get_page)
        d = pinger.ping("http://ping/url", None)
        d.addCallback(self.assertEqual, False)
        self.assertEqual(client.fetches, [])

    def test_respond(self):
        """
        The L{PingClient.ping} fire the Deferred it returns with True if the
        web request indicates that the computer has messages.
        """
        client = FakePageGetter({"messages": True})
        pinger = PingClient(self.reactor, get_page=client.get_page)
        d = pinger.ping("http://ping/url", 23)
        d.addCallback(self.assertEqual, True)

    def test_errback(self):
        """
        If the HTTP request fails the deferred returned by L{PingClient.ping}
        fires back with an error.
        """
        client = FakePageGetter(None)
        pinger = PingClient(self.reactor, get_page=client.failing_get_page)
        d = pinger.ping("http://ping/url", 23)
        failures = []

        def errback(failure):
            failures.append(failure)
        d.addErrback(errback)
        self.assertEqual(len(failures), 1)
        self.assertEqual(failures[0].getErrorMessage(), "That's a failure!")
        self.assertEqual(failures[0].type, AssertionError)


class PingerTest(LandscapeTest):

    helpers = [ExchangeHelper]

    # Tell the Plugin helper to not add a MessageExchange plugin, to interfere
    # with our code which asserts stuff about when *our* plugin fires
    # exchanges.
    install_exchanger = False

    def setUp(self):
        super(PingerTest, self).setUp()
        self.page_getter = FakePageGetter(None)

        def factory(reactor):
            return PingClient(reactor, get_page=self.page_getter.get_page)

        self.config.ping_url = "http://localhost:8081/whatever"
        self.config.ping_interval = 10

        self.pinger = Pinger(self.reactor,
                             self.identity,
                             self.exchanger,
                             self.config,
                             ping_client_factory=factory)

    def test_default_ping_client(self):
        """
        The C{ping_client_factory} argument to L{Pinger} should be optional,
        and default to L{PingClient}.
        """
        pinger = Pinger(self.reactor,
                        self.identity,
                        self.exchanger,
                        self.config)
        self.assertEqual(pinger.ping_client_factory, PingClient)

    def test_occasional_ping(self):
        """
        The L{Pinger} should be able to occasionally ask if there are
        messages.
        """
        self.pinger.start()
        self.identity.insecure_id = 23
        self.reactor.advance(9)
        self.assertEqual(len(self.page_getter.fetches), 0)
        self.reactor.advance(1)
        self.assertEqual(len(self.page_getter.fetches), 1)

    def test_load_insecure_id(self):
        """
        If the insecure-id has already been saved when the plugin is
        registered, it should immediately start pinging.
        """
        self.identity.insecure_id = 42
        self.pinger.start()
        self.reactor.advance(10)
        self.assertEqual(len(self.page_getter.fetches), 1)

    def test_response(self):
        """
        When a ping indicates there are messages, an exchange should occur.
        """
        self.pinger.start()
        self.identity.insecure_id = 42
        self.page_getter.response = {"messages": True}

        # 70 = ping delay + urgent exchange delay
        self.reactor.advance(70)

        self.assertEqual(len(self.transport.payloads), 1)

    def test_negative_response(self):
        """
        When a ping indicates there are no messages, no exchange should occur.
        """
        self.pinger.start()
        self.identity.insecure_id = 42
        self.page_getter.response = {"messages": False}
        self.reactor.advance(10)
        self.assertEqual(len(self.transport.payloads), 0)

    def test_ping_error(self):
        """
        When the web interaction fails for some reason, a message
        should be logged.
        """
        self.log_helper.ignore_errors(ZeroDivisionError)
        self.identity.insecure_id = 42

        class BadPingClient(object):
            def __init__(self, *args, **kwargs):
                pass

            def ping(self, url, secure_id):
                self.url = url
                return fail(ZeroDivisionError("Couldn't fetch page"))

        self.config.ping_url = "http://foo.com/"
        pinger = Pinger(self.reactor,
                        self.identity,
                        self.exchanger,
                        self.config,
                        ping_client_factory=BadPingClient)
        pinger.start()

        self.reactor.advance(30)

        log = self.logfile.getvalue()
        self.assertTrue("Error contacting ping server at "
                        "http://foo.com/" in log,
                        log)
        self.assertTrue("ZeroDivisionError" in log)
        self.assertTrue("Couldn't fetch page" in log)

    def test_get_interval(self):
        self.assertEqual(self.pinger.get_interval(), 10)

    def test_set_intervals_handling(self):
        self.pinger.start()

        self.reactor.fire("message", {"type": "set-intervals", "ping": 73})
        self.assertEqual(self.pinger.get_interval(), 73)

        # The server may set specific intervals only, not including the ping.
        self.reactor.fire("message", {"type": "set-intervals"})
        self.assertEqual(self.pinger.get_interval(), 73)

        self.identity.insecure_id = 23
        self.reactor.advance(72)
        self.assertEqual(len(self.page_getter.fetches), 0)
        self.reactor.advance(1)
        self.assertEqual(len(self.page_getter.fetches), 1)

    def test_get_url(self):
        self.assertEqual(self.pinger.get_url(),
                         "http://localhost:8081/whatever")

    def test_config_url(self):
        """
        The L{Pinger} uses the ping URL set in the given configuration.
        """
        self.identity.insecure_id = 23
        url = "http://example.com/mysuperping"
        self.config.ping_url = url
        self.pinger.start()
        self.reactor.advance(10)
        self.assertEqual(self.page_getter.fetches[0][0], url)

    def test_reschedule(self):
        """
        Each time a ping is completed the L{Pinger} schedules a new ping using
        the current ping interval.
        """
        self.identity.insecure_id = 23
        self.pinger.start()
        self.reactor.advance(10)
        self.assertEqual(1, len(self.page_getter.fetches))
        self.reactor.advance(10)
        self.assertEqual(2, len(self.page_getter.fetches))

    def test_reschedule_with_ping_interval_change(self):
        """
        If the ping interval changes, new pings will be scheduled accordingly.
        """
        self.identity.insecure_id = 23
        self.pinger.start()
        self.reactor.advance(5)
        # Simulate interval changing in the meantime
        self.config.ping_interval = 20
        self.reactor.advance(5)
        self.assertEqual(1, len(self.page_getter.fetches))
        # The new interval is 20, so after only 10 seconds nothing happens
        self.reactor.advance(10)
        self.assertEqual(1, len(self.page_getter.fetches))
        # After another 10 seconds we reach the 20 seconds interval and the
        # ping is triggered
        self.reactor.advance(10)
        self.assertEqual(2, len(self.page_getter.fetches))

    def test_change_url_after_start(self):
        """
        If the C{ping_url} set in the configuration is changed after the
        pinger has started, the target HTTP url will adjust accordingly.
        """
        url = "http://example.com/mysuperping"
        self.pinger.start()
        self.config.ping_url = url
        self.identity.insecure_id = 23
        self.reactor.advance(10)
        self.assertEqual(self.page_getter.fetches[0][0], url)

    def test_ping_doesnt_ping_if_stopped(self):
        """If the L{Pinger} is stopped, no pings are performed."""
        self.pinger.start()
        self.pinger.stop()
        self.reactor.advance(10)
        self.assertEqual([], self.page_getter.fetches)



from landscape.tests.helpers import LandscapeTest

from landscape.schema import (
    InvalidError, Constant, Bool, Int, Float, Bytes, Unicode, List, KeyDict,
    Dict, Tuple, Any, Message)


class DummySchema(object):
    def coerce(self, value):
        return "hello!"


class BasicTypesTest(LandscapeTest):

    def test_any(self):
        schema = Any(Constant(None), Unicode())
        self.assertEqual(schema.coerce(None), None)
        self.assertEqual(schema.coerce(u"foo"), u"foo")

    def test_any_bad(self):
        schema = Any(Constant(None), Unicode())
        self.assertRaises(InvalidError, schema.coerce, object())

    def test_constant(self):
        self.assertEqual(Constant("hello").coerce("hello"), "hello")

    def test_constant_arbitrary(self):
        obj = object()
        self.assertEqual(Constant(obj).coerce(obj), obj)

    def test_constant_bad(self):
        self.assertRaises(InvalidError, Constant("foo").coerce, object())

    def test_bool(self):
        self.assertEqual(Bool().coerce(True), True)
        self.assertEqual(Bool().coerce(False), False)

    def test_bool_bad(self):
        self.assertRaises(InvalidError, Bool().coerce, 1)

    def test_int(self):
        self.assertEqual(Int().coerce(3), 3)

    def test_int_accepts_long(self):
        self.assertEqual(Int().coerce(3L), 3)

    def test_int_bad_str(self):
        self.assertRaises(InvalidError, Int().coerce, "3")

    def test_int_bad_float(self):
        self.assertRaises(InvalidError, Int().coerce, 3.0)

    def test_float(self):
        self.assertEqual(Float().coerce(3.3), 3.3)

    def test_float_accepts_int(self):
        self.assertEqual(Float().coerce(3), 3.0)

    def test_float_accepts_long(self):
        self.assertEqual(Float().coerce(3L), 3.0)

    def test_float_bad_str(self):
        self.assertRaises(InvalidError, Float().coerce, "3.0")

    def test_string(self):
        self.assertEqual(Bytes().coerce("foo"), "foo")

    def test_string_bad_unicode(self):
        self.assertRaises(InvalidError, Bytes().coerce, u"foo")

    def test_string_bad_anything(self):
        self.assertRaises(InvalidError, Bytes().coerce, object())

    def test_unicode(self):
        self.assertEqual(Unicode().coerce(u"foo"), u"foo")

    def test_unicode_bad_value(self):
        """Invalid values raise an errors."""
        self.assertRaises(InvalidError, Unicode().coerce, 32)

    def test_unicode_with_str(self):
        """Unicode accept plain strings and return a unicode."""
        self.assertEqual(Unicode().coerce("foo"), u"foo")

    def test_unicode_decodes(self):
        """Unicode should decode plain strings."""
        a = u"\N{HIRAGANA LETTER A}"
        self.assertEqual(Unicode().coerce(a.encode("utf-8")), a)
        letter = u"\N{LATIN SMALL LETTER A WITH GRAVE}"
        self.assertEqual(
            Unicode(encoding="latin-1").coerce(letter.encode("latin-1")),
            letter)

    def test_unicode_or_str_bad_encoding(self):
        """Decoding errors should be converted to InvalidErrors."""
        self.assertRaises(InvalidError, Unicode().coerce, "\xff")

    def test_list(self):
        self.assertEqual(List(Int()).coerce([1]), [1])

    def test_list_bad(self):
        self.assertRaises(InvalidError, List(Int()).coerce, 32)

    def test_list_inner_schema_coerces(self):
        self.assertEqual(List(DummySchema()).coerce([3]), ["hello!"])

    def test_list_bad_inner_schema(self):
        self.assertRaises(InvalidError, List(Int()).coerce, ["hello"])

    def test_list_multiple_items(self):
        a = u"\N{HIRAGANA LETTER A}"
        schema = List(Unicode())
        self.assertEqual(schema.coerce([a, a.encode("utf-8")]), [a, a])

    def test_tuple(self):
        self.assertEqual(Tuple(Int()).coerce((1,)), (1,))

    def test_tuple_coerces(self):
        self.assertEqual(Tuple(Int(), DummySchema()).coerce((23, object())),
                         (23, "hello!"))

    def test_tuple_bad(self):
        self.assertRaises(InvalidError, Tuple().coerce, object())

    def test_tuple_inner_schema_bad(self):
        self.assertRaises(InvalidError, Tuple(Int()).coerce, (object(),))

    def test_tuple_must_have_all_items(self):
        self.assertRaises(InvalidError, Tuple(Int(), Int()).coerce, (1,))

    def test_tuple_must_have_no_more_items(self):
        self.assertRaises(InvalidError, Tuple(Int()).coerce, (1, 2))

    def test_key_dict(self):
        self.assertEqual(KeyDict({"foo": Int()}).coerce({"foo": 1}),
                         {"foo": 1})

    def test_key_dict_coerces(self):
        self.assertEqual(KeyDict({"foo": DummySchema()}).coerce({"foo": 3}),
                         {"foo": "hello!"})

    def test_key_dict_bad_inner_schema(self):
        self.assertRaises(InvalidError, KeyDict({"foo": Int()}).coerce,
                          {"foo": "hello"})

    def test_key_dict_unknown_key(self):
        self.assertRaises(InvalidError, KeyDict({}).coerce, {"foo": 1})

    def test_key_dict_bad(self):
        self.assertRaises(InvalidError, KeyDict({}).coerce, object())

    def test_key_dict_multiple_items(self):
        schema = KeyDict({"one": Int(), "two": List(Float())})
        input = {"one": 32, "two": [1.5, 2.3]}
        self.assertEqual(schema.coerce(input),
                         {"one": 32, "two": [1.5, 2.3]})

    def test_key_dict_arbitrary_keys(self):
        """
        KeyDict doesn't actually need to have strings as keys, just any
        object which hashes the same.
        """
        key = object()
        self.assertEqual(KeyDict({key: Int()}).coerce({key: 32}), {key: 32})

    def test_key_dict_must_have_all_keys(self):
        """
        dicts which are applied to a KeyDict must have all the keys
        specified in the KeyDict.
        """
        schema = KeyDict({"foo": Int()})
        self.assertRaises(InvalidError, schema.coerce, {})

    def test_key_dict_optional_keys(self):
        """KeyDict allows certain keys to be optional.
        """
        schema = KeyDict({"foo": Int(), "bar": Int()}, optional=["bar"])
        self.assertEqual(schema.coerce({"foo": 32}), {"foo": 32})

    def test_pass_optional_key(self):
        """Regression test. It should be possible to pass an optional key.
        """
        schema = KeyDict({"foo": Int()}, optional=["foo"])
        self.assertEqual(schema.coerce({"foo": 32}), {"foo": 32})

    def test_dict(self):
        self.assertEqual(Dict(Int(), Bytes()).coerce({32: "hello."}),
                         {32: "hello."})

    def test_dict_coerces(self):
        self.assertEqual(
            Dict(DummySchema(), DummySchema()).coerce({32: object()}),
            {"hello!": "hello!"})

    def test_dict_inner_bad(self):
        self.assertRaises(InvalidError, Dict(Int(), Int()).coerce, {"32": 32})

    def test_dict_wrong_type(self):
        self.assertRaises(InvalidError, Dict(Int(), Int()).coerce, 32)

    def test_message(self):
        """The L{Message} schema should be very similar to KeyDict."""
        schema = Message("foo", {"data": Int()})
        self.assertEqual(
            schema.coerce({"type": "foo", "data": 3}),
            {"type": "foo", "data": 3})

    def test_message_timestamp(self):
        """L{Message} schemas should accept C{timestamp} keys."""
        schema = Message("bar", {})
        self.assertEqual(
            schema.coerce({"type": "bar", "timestamp": 0.33}),
            {"type": "bar", "timestamp": 0.33})

    def test_message_api(self):
        """L{Message} schemas should accept C{api} keys."""
        schema = Message("baz", {})
        self.assertEqual(
            schema.coerce({"type": "baz", "api": "whatever"}),
            {"type": "baz", "api": "whatever"})

    def test_message_api_None(self):
        """L{Message} schemas should accept None for C{api}."""
        schema = Message("baz", {})
        self.assertEqual(
            schema.coerce({"type": "baz", "api": None}),
            {"type": "baz", "api": None})

    def test_message_optional(self):
        """The L{Message} schema should allow additional optional keys."""
        schema = Message("foo", {"data": Int()}, optional=["data"])
        self.assertEqual(schema.coerce({"type": "foo"}), {"type": "foo"})

    def test_message_type(self):
        """The C{type} should be introspectable on L{Message} objects."""
        schema = Message("foo", {})
        self.assertEqual(schema.type, "foo")

    def test_message_with_unknown_fields(self):
        """
        The L{Message} schema discards unknown fields when coercing values.
        """
        schema = Message("foo", {})
        self.assertEqual({"type": "foo"},
                         schema.coerce({"type": "foo", "crap": 123}))



from landscape.tests.helpers import LandscapeTest
from landscape.monitor.config import MonitorConfiguration, ALL_PLUGINS


class MonitorConfigurationTest(LandscapeTest):

    def setUp(self):
        super(MonitorConfigurationTest, self).setUp()
        self.config = MonitorConfiguration()

    def test_plugin_factories(self):
        """
        By default all plugins are enabled.
        """
        self.assertEqual(self.config.plugin_factories, ALL_PLUGINS)

    def test_plugin_factories_with_monitor_plugins(self):
        """
        The C{--monitor-plugins} command line option can be used to specify
        which plugins should be active.
        """
        self.config.load(["--monitor-plugins", "  ComputerInfo, LoadAverage "])
        self.assertEqual(
            self.config.plugin_factories, ["ComputerInfo", "LoadAverage"])

    def test_flush_interval(self):
        """
        The C{--flush-interval} command line option can be used to specify the
        flush interval.
        """
        self.config.load(["--flush-interval", "123"])
        self.assertEqual(self.config.flush_interval, 123)
